import numpy as np
from functools import wraps

import autofit as af
from autoarray.structures import grids
from autoastro import dimensions as dim


def transform_grid(func):
    """Wrap the function in a function that checks whether the coordinates have been transformed. If they have not \ 
    been transformed then they are transformed.

    Parameters
    ----------
    func : (profiles, *args, **kwargs) -> Object
        A function that requires transformed coordinates

    Returns
    -------
        A function that can except cartesian or transformed coordinates
    """

    @wraps(func)
    def wrapper(profile, grid, *args, **kwargs):
        """

        Parameters
        ----------
        profile : GeometryProfile
            The profiles that owns the function
        grid : ndarray
            PlaneCoordinates in either cartesian or profiles coordinate system
        args
        kwargs

        Returns
        -------
            A value or coordinate in the same coordinate system as those passed in.
        """

        if not isinstance(grid, TransformedGrid):
            result = func(
                profile,
                profile.transform_grid_to_reference_frame(grid),
                *args,
                **kwargs,
            )

            if len(result.shape) == 1:
                return grid.mapping.array_stored_1d_from_sub_array_1d(
                    sub_array_1d=result
                )
            else:
                return grid.mapping.grid_stored_1d_from_sub_grid_1d(sub_grid_1d=result)

        else:
            return func(profile, grid, *args, **kwargs)

    return wrapper


def cache(func):
    """
    Caches results of a call to a grid function. If a grid that evaluates to the same byte value is passed into the same
    function of the same instance as previously then the cached result is returned.

    Parameters
    ----------
    func
        Some instance method that takes a grid as its argument

    Returns
    -------
    result
        Some result, either newly calculated or recovered from the cache
    """

    def wrapper(instance: GeometryProfile, grid: np.ndarray, *args, **kwargs):
        if not hasattr(instance, "cache"):
            instance.cache = {}
        key = (func.__name__, grid.tobytes())
        if key not in instance.cache:
            instance.cache[key] = func(instance, grid)
        return instance.cache[key]

    return wrapper


def move_grid_to_radial_minimum(func):
    """ Checks whether any coordinates in the grid are radially near (0.0, 0.0), which can lead to numerical faults in \
    the evaluation of a light or mass profiles. If any coordinates are radially within the the radial minimum \
    threshold, their (y,x) coordinates are shifted to that value to ensure they are evaluated correctly.

    By default this radial minimum is not used, and users should be certain they use a value that does not impact \
    results.

    Parameters
    ----------
    func : (profile, *args, **kwargs) -> Object
        A function that takes a grid of coordinates which may have a singularity as (0.0, 0.0)

    Returns
    -------
        A function that can except cartesian or transformed coordinates
    """

    @wraps(func)
    def wrapper(profile, grid, *args, **kwargs):
        """

        Parameters
        ----------
        profile : SphericalProfile
            The profiles that owns the function
        grid : ndarray
            PlaneCoordinates in either cartesian or profiles coordinate system
        args
        kwargs

        Returns
        -------
            A value or coordinate in the same coordinate system as those passed in.
        """
        radial_minimum_config = af.conf.NamedConfig(
            f"{af.conf.instance.config_path}/radial_minimum.ini"
        )
        grid_radial_minimum = radial_minimum_config.get(
            "radial_minimum", profile.__class__.__name__, float
        )
        with np.errstate(all="ignore"):  # Division by zero fixed via isnan
            grid_radii = profile.grid_to_grid_radii(grid=grid)
            grid_radial_scale = np.where(
                grid_radii < grid_radial_minimum, grid_radial_minimum / grid_radii, 1.0
            )
            grid = np.multiply(grid, grid_radial_scale[:, None])
        grid[np.isnan(grid)] = grid_radial_minimum
        return func(profile, grid, *args, **kwargs)

    return wrapper


class TransformedGrid(grids.AbstractGrid):
    pass


class GeometryProfile(dim.DimensionsProfile):
    @af.map_types
    def __init__(self, centre: dim.Position = (0.0, 0.0)):
        """An abstract geometry profile, which describes profiles with y and x centre Cartesian coordinates
        
        Parameters
        -----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        """

        self.centre = centre

    def transform_grid_to_reference_frame(self, grid):
        raise NotImplemented()

    def transform_grid_from_reference_frame(self, grid):
        raise NotImplemented()

    def __repr__(self):
        return "{}\n{}".format(
            self.__class__.__name__,
            "\n".join(["{}: {}".format(k, v) for k, v in self.__dict__.items()]),
        )

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class SphericalProfile(GeometryProfile):
    @af.map_types
    def __init__(self, centre: dim.Position = (0.0, 0.0)):
        """ A spherical profile, which describes profiles with y and x centre Cartesian coordinates.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        """
        super(SphericalProfile, self).__init__(centre=centre)

    @grids.convert_coordinates_to_grid
    @transform_grid
    def grid_to_grid_radii(self, grid):
        """Convert a grid of (y, x) coordinates to a grid of their circular radii.

        If the coordinates have not been transformed to the profile's centre, this is performed automatically.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the profile.
        """
        return np.sqrt(np.add(np.square(grid[:, 0]), np.square(grid[:, 1])))

    def grid_angle_to_profile(self, grid_thetas):
        """The angle between each (y,x) coordinate on the grid and the profile, in radians.
        
        Parameters
        -----------
        grid_thetas : ndarray
            The angle theta counter-clockwise from the positive x-axis to each coordinate in radians.
        """
        return np.cos(grid_thetas), np.sin(grid_thetas)

    def grid_to_grid_cartesian(self, grid, radius):
        """
        Convert a grid of (y,x) coordinates with their specified circular radii to their original (y,x) Cartesian 
        coordinates.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the profile.
        radius : ndarray
            The circular radius of each coordinate from the profile center.
        """
        grid_thetas = np.arctan2(grid[:, 0], grid[:, 1])
        cos_theta, sin_theta = self.grid_angle_to_profile(grid_thetas=grid_thetas)
        return np.multiply(radius[:, None], np.vstack((sin_theta, cos_theta)).T)

    def transform_grid_to_reference_frame(self, grid):
        """Transform a grid of (y,x) coordinates to the reference frame of the profile, including a translation to \
        its centre.

        Parameters
        ----------
        grid : ndarray
            The (y, x) coordinates in the original reference frame of the grid.
        """
        transformed = np.subtract(grid, self.centre)
        return TransformedGrid(grid=transformed, mask=grid.mask)

    def transform_grid_from_reference_frame(self, grid):
        """Transform a grid of (y,x) coordinates from the reference frame of the profile to the original observer \
        reference frame, including a translation from the profile's centre.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the profile.
        """
        transformed = np.add(grid, self.centre)
        return transformed.view(TransformedGrid)


class EllipticalProfile(SphericalProfile):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        axis_ratio: float = 1.0,
        phi: float = 0.0,
    ):
        """ An elliptical profile, which describes profiles with y and x centre Cartesian coordinates, an axis-ratio \
        and rotational angle phi.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotation angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalProfile, self).__init__(centre=centre)
        self.axis_ratio = axis_ratio
        self.phi = phi

    @property
    def phi_radians(self):
        return np.radians(self.phi)

    @property
    def cos_phi(self):
        return self.cos_and_sin_from_x_axis()[0]

    @property
    def sin_phi(self):
        return self.cos_and_sin_from_x_axis()[1]

    def cos_and_sin_from_x_axis(self):
        """ Determine the sin and cosine of the angle between the profile's ellipse and the positive x-axis, \
        counter-clockwise. """
        phi_radians = np.radians(self.phi)
        return np.cos(phi_radians), np.sin(phi_radians)

    def grid_angle_to_profile(self, grid_thetas):
        """The angle between each angle theta on the grid and the profile, in radians.

        Parameters
        -----------
        grid_thetas : ndarray
            The angle theta counter-clockwise from the positive x-axis to each coordinate in radians.
        """
        theta_coordinate_to_profile = np.add(grid_thetas, -self.phi_radians)
        return np.cos(theta_coordinate_to_profile), np.sin(theta_coordinate_to_profile)

    def rotate_grid_from_profile(self, grid_elliptical):
        """ Rotate a grid of elliptical (y,x) coordinates from the reference frame of the profile back to the \
        unrotated coordinate grid reference frame (coordinates are not shifted back to their original centre).

        This routine is used after computing deflection angles in the reference frame of the profile, so that the \
        deflection angles can be re-rotated to the frame of the original coordinates before performing ray-tracing.

        Parameters
        ----------
        grid_elliptical : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of an elliptical profile.
        """
        y = np.add(
            np.multiply(grid_elliptical[:, 1], self.sin_phi),
            np.multiply(grid_elliptical[:, 0], self.cos_phi),
        )
        x = np.add(
            np.multiply(grid_elliptical[:, 1], self.cos_phi),
            -np.multiply(grid_elliptical[:, 0], self.sin_phi),
        )
        return np.vstack((y, x)).T

    @grids.convert_coordinates_to_grid
    @transform_grid
    @move_grid_to_radial_minimum
    def grid_to_elliptical_radii(self, grid):
        """ Convert a grid of (y,x) coordinates to an elliptical radius.

        If the coordinates have not been transformed to the profile's geometry, this is performed automatically.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the elliptical profile.
        """
        return np.sqrt(
            np.add(
                np.square(grid[:, 1]), np.square(np.divide(grid[:, 0], self.axis_ratio))
            )
        )

    @grids.convert_coordinates_to_grid
    @transform_grid
    @move_grid_to_radial_minimum
    def grid_to_eccentric_radii(self, grid):
        """Convert a grid of (y,x) coordinates to an eccentric radius, which is (1.0/axis_ratio) * elliptical radius \
        and used to define light profile half-light radii using circular radii.

        If the coordinates have not been transformed to the profile's geometry, this is performed automatically.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the elliptical profile.
        """
        return np.multiply(
            np.sqrt(self.axis_ratio), self.grid_to_elliptical_radii(grid)
        ).view(np.ndarray)

    def transform_grid_to_reference_frame(self, grid):
        """Transform a grid of (y,x) coordinates to the reference frame of the profile, including a translation to \
        its centre and a rotation to it orientation.

        Parameters
        ----------
        grid : ndarray
            The (y, x) coordinates in the original reference frame of the grid.
        """
        if self.__class__.__name__.startswith("Spherical"):
            return super().transform_grid_to_reference_frame(grid)
        shifted_coordinates = np.subtract(grid, self.centre)
        radius = np.sqrt(np.sum(shifted_coordinates ** 2.0, 1))
        theta_coordinate_to_profile = (
            np.arctan2(shifted_coordinates[:, 0], shifted_coordinates[:, 1])
            - self.phi_radians
        )
        transformed = np.vstack(
            (
                radius * np.sin(theta_coordinate_to_profile),
                radius * np.cos(theta_coordinate_to_profile),
            )
        ).T
        return TransformedGrid(grid=transformed, mask=grid.mask)

    def transform_grid_from_reference_frame(self, grid):
        """Transform a grid of (y,x) coordinates from the reference frame of the profile to the original observer \
        reference frame, including a rotation to its original orientation and a translation from the profile's centre.

        Parameters
        ----------
        grid : TransformedGrid(ndarray)
            The (y, x) coordinates in the reference frame of the profile.
        """
        if self.__class__.__name__.startswith("Spherical"):
            return super().transform_grid_from_reference_frame(grid)

        y = np.add(
            np.add(
                np.multiply(grid[:, 1], self.sin_phi),
                np.multiply(grid[:, 0], self.cos_phi),
            ),
            self.centre[0],
        )
        x = np.add(
            np.add(
                np.multiply(grid[:, 1], self.cos_phi),
                -np.multiply(grid[:, 0], self.sin_phi),
            ),
            self.centre[1],
        )
        return np.vstack((y, x)).T

    def eta_u(self, u, coordinates):
        return np.sqrt(
            (
                u
                * (
                    (coordinates[1] ** 2)
                    + (coordinates[0] ** 2 / (1 - (1 - self.axis_ratio ** 2) * u))
                )
            )
        )



from autoastro import exc
import pytest

import autoastro as aast

from test_autoastro.mock import mock_cosmology


class TestLength:
    def test__conversions_from_arcsec_to_kpc_and_back__errors_raised_if_no_kpc_per_arcsec(
        self
    ):
        unit_arcsec = aast.dim.Length(value=2.0)

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit_length == "arcsec"

        unit_arcsec = unit_arcsec.convert(unit_length="arcsec")

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == "arcsec"

        unit_kpc = unit_arcsec.convert(unit_length="kpc", kpc_per_arcsec=2.0)

        assert unit_kpc == 4.0
        assert unit_kpc.unit == "kpc"

        unit_kpc = unit_kpc.convert(unit_length="kpc")

        assert unit_kpc == 4.0
        assert unit_kpc.unit == "kpc"

        unit_arcsec = unit_kpc.convert(unit_length="arcsec", kpc_per_arcsec=2.0)

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == "arcsec"

        with pytest.raises(exc.UnitsException):
            unit_arcsec.convert(unit_length="kpc")
            unit_kpc.convert(unit_length="arcsec")
            unit_arcsec.convert(unit_length="lol")


class TestLuminosity:
    def test__conversions_from_eps_and_counts_and_back__errors_raised_if_no_exposure_time(
        self
    ):

        unit_eps = aast.dim.Luminosity(value=2.0)

        assert unit_eps == 2.0
        assert unit_eps.unit_luminosity == "eps"

        unit_eps = unit_eps.convert(unit_luminosity="eps")

        assert unit_eps == 2.0
        assert unit_eps.unit == "eps"

        unit_counts = unit_eps.convert(unit_luminosity="counts", exposure_time=2.0)

        assert unit_counts == 4.0
        assert unit_counts.unit == "counts"

        unit_counts = unit_counts.convert(unit_luminosity="counts")

        assert unit_counts == 4.0
        assert unit_counts.unit == "counts"

        unit_eps = unit_counts.convert(unit_luminosity="eps", exposure_time=2.0)

        assert unit_eps == 2.0
        assert unit_eps.unit == "eps"

        with pytest.raises(exc.UnitsException):
            unit_eps.convert(unit_luminosity="counts")
            unit_counts.convert(unit_luminosity="eps")
            unit_eps.convert(unit_luminosity="lol")


class TestMass:
    def test__conversions_from_angular_and_sol_mass_and_back__errors_raised_if_no_exposure_time(
        self
    ):

        mass_angular = aast.dim.Mass(value=2.0)

        assert mass_angular == 2.0
        assert mass_angular.unit_mass == "angular"

        # angular -> angular, stays 2.0

        mass_angular = mass_angular.convert(unit_mass="angular")

        assert mass_angular == 2.0
        assert mass_angular.unit == "angular"

        # angular -> solMass, converts to 2.0 * 2.0 = 4.0

        mas_sol_mass = mass_angular.convert(
            unit_mass="solMass", critical_surface_density=2.0
        )

        assert mas_sol_mass == 4.0
        assert mas_sol_mass.unit == "solMass"

        # solMass -> solMass, stays 4.0

        mas_sol_mass = mas_sol_mass.convert(unit_mass="solMass")

        assert mas_sol_mass == 4.0
        assert mas_sol_mass.unit == "solMass"

        # solMass -> angular, stays 4.0

        mass_angular = mas_sol_mass.convert(
            unit_mass="angular", critical_surface_density=2.0
        )

        assert mass_angular == 2.0
        assert mass_angular.unit == "angular"

        with pytest.raises(exc.UnitsException):
            mass_angular.convert(unit_mass="solMass")
            mas_sol_mass.convert(unit_mass="angular")
            mass_angular.convert(unit_mass="lol")


class TestMassOverLuminosity:
    def test__conversions_from_angular_and_sol_mass_and_back__errors_raised_if_critical_mass_density(
        self
    ):

        unit_angular = aast.dim.MassOverLuminosity(value=2.0)

        assert unit_angular == 2.0
        assert unit_angular.unit == "angular / eps"

        unit_angular = unit_angular.convert(unit_mass="angular", unit_luminosity="eps")

        assert unit_angular == 2.0
        assert unit_angular.unit == "angular / eps"

        unit_sol_mass = unit_angular.convert(
            unit_mass="solMass", critical_surface_density=2.0, unit_luminosity="eps"
        )

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == "solMass / eps"

        unit_sol_mass = unit_sol_mass.convert(
            unit_mass="solMass", unit_luminosity="eps"
        )

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == "solMass / eps"

        unit_angular = unit_sol_mass.convert(
            unit_mass="angular", critical_surface_density=2.0, unit_luminosity="eps"
        )

        assert unit_angular == 2.0
        assert unit_angular.unit == "angular / eps"

        with pytest.raises(exc.UnitsException):
            unit_angular.convert(unit_mass="solMass", unit_luminosity="eps")
            unit_sol_mass.convert(unit_mass="angular", unit_luminosity="eps")
            unit_angular.convert(unit_mass="lol", unit_luminosity="eps")

    def test__conversions_from_eps_and_counts_and_back__errors_raised_if_no_exposure_time(
        self
    ):

        unit_eps = aast.dim.MassOverLuminosity(value=2.0)

        assert unit_eps == 2.0
        assert unit_eps.unit == "angular / eps"

        unit_eps = unit_eps.convert(unit_mass="angular", unit_luminosity="eps")

        assert unit_eps == 2.0
        assert unit_eps.unit == "angular / eps"

        unit_counts = unit_eps.convert(
            unit_mass="angular", exposure_time=2.0, unit_luminosity="counts"
        )

        assert unit_counts == 1.0
        assert unit_counts.unit == "angular / counts"

        unit_counts = unit_counts.convert(unit_mass="angular", unit_luminosity="counts")

        assert unit_counts == 1.0
        assert unit_counts.unit == "angular / counts"

        unit_eps = unit_counts.convert(
            unit_mass="angular", exposure_time=2.0, unit_luminosity="eps"
        )

        assert unit_eps == 2.0
        assert unit_eps.unit == "angular / eps"

        with pytest.raises(exc.UnitsException):
            unit_eps.convert(unit_mass="angular", unit_luminosity="eps")
            unit_counts.convert(unit_mass="angular", unit_luminosity="eps")
            unit_eps.convert(unit_mass="lol", unit_luminosity="eps")


class TestMassOverLength2:
    def test__conversions_from_angular_and_sol_mass_and_back__errors_raised_if_critical_mass_density(
        self
    ):

        unit_angular = aast.dim.MassOverLength2(value=2.0)

        assert unit_angular == 2.0
        assert unit_angular.unit == "angular / arcsec^2"

        unit_angular = unit_angular.convert(unit_mass="angular", unit_length="arcsec")

        assert unit_angular == 2.0
        assert unit_angular.unit == "angular / arcsec^2"

        unit_sol_mass = unit_angular.convert(
            unit_mass="solMass", critical_surface_density=2.0, unit_length="arcsec"
        )

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == "solMass / arcsec^2"

        unit_sol_mass = unit_sol_mass.convert(unit_mass="solMass", unit_length="arcsec")

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == "solMass / arcsec^2"

        unit_angular = unit_sol_mass.convert(
            unit_mass="angular", critical_surface_density=2.0, unit_length="arcsec"
        )

        assert unit_angular == 2.0
        assert unit_angular.unit == "angular / arcsec^2"

        with pytest.raises(exc.UnitsException):
            unit_angular.convert(unit_mass="solMass", unit_length="eps")
            unit_sol_mass.convert(unit_mass="angular", unit_length="eps")
            unit_angular.convert(unit_mass="lol", unit_length="eps")

    def test__conversions_from_arcsec_to_kpc_and_back__errors_raised_if_no_kpc_per_arcsec(
        self
    ):

        unit_arcsec = aast.dim.MassOverLength2(value=2.0, unit_mass="solMass")

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == "solMass / arcsec^2"

        unit_arcsec = unit_arcsec.convert(unit_length="arcsec", unit_mass="solMass")

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == "solMass / arcsec^2"

        unit_kpc = unit_arcsec.convert(
            unit_length="kpc", kpc_per_arcsec=2.0, unit_mass="solMass"
        )

        assert unit_kpc == 2.0 / 2.0 ** 2.0
        assert unit_kpc.unit == "solMass / kpc^2"

        unit_kpc = unit_kpc.convert(unit_length="kpc", unit_mass="solMass")

        assert unit_kpc == 2.0 / 2.0 ** 2.0
        assert unit_kpc.unit == "solMass / kpc^2"

        unit_arcsec = unit_kpc.convert(
            unit_length="arcsec", kpc_per_arcsec=2.0, unit_mass="solMass"
        )

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == "solMass / arcsec^2"

        with pytest.raises(exc.UnitsException):
            unit_arcsec.convert(unit_length="kpc", unit_mass="solMass")
            unit_kpc.convert(unit_length="arcsec", unit_mass="solMass")
            unit_arcsec.convert(unit_length="lol", unit_mass="solMass")


class TestMassOverLength3:
    def test__conversions_from_angular_and_sol_mass_and_back__errors_raised_if_critical_mass_density(
        self
    ):

        unit_angular = aast.dim.MassOverLength3(value=2.0)

        assert unit_angular == 2.0
        assert unit_angular.unit == "angular / arcsec^3"

        unit_angular = unit_angular.convert(unit_mass="angular", unit_length="arcsec")

        assert unit_angular == 2.0
        assert unit_angular.unit == "angular / arcsec^3"

        unit_sol_mass = unit_angular.convert(
            unit_mass="solMass", critical_surface_density=2.0, unit_length="arcsec"
        )

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == "solMass / arcsec^3"

        unit_sol_mass = unit_sol_mass.convert(unit_mass="solMass", unit_length="arcsec")

        assert unit_sol_mass == 4.0
        assert unit_sol_mass.unit == "solMass / arcsec^3"

        unit_angular = unit_sol_mass.convert(
            unit_mass="angular", critical_surface_density=2.0, unit_length="arcsec"
        )

        assert unit_angular == 2.0
        assert unit_angular.unit == "angular / arcsec^3"

        with pytest.raises(exc.UnitsException):
            unit_angular.convert(unit_mass="solMass", unit_length="eps")
            unit_sol_mass.convert(unit_mass="angular", unit_length="eps")
            unit_angular.convert(unit_mass="lol", unit_length="eps")

    def test__conversions_from_arcsec_to_kpc_and_back__errors_raised_if_no_kpc_per_arcsec(
        self
    ):

        unit_arcsec = aast.dim.MassOverLength3(value=2.0, unit_mass="solMass")

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == "solMass / arcsec^3"

        unit_arcsec = unit_arcsec.convert(unit_length="arcsec", unit_mass="solMass")

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == "solMass / arcsec^3"

        unit_kpc = unit_arcsec.convert(
            unit_length="kpc", kpc_per_arcsec=2.0, unit_mass="solMass"
        )

        assert unit_kpc == 2.0 / 2.0 ** 3.0
        assert unit_kpc.unit == "solMass / kpc^3"

        unit_kpc = unit_kpc.convert(unit_length="kpc", unit_mass="solMass")

        assert unit_kpc == 2.0 / 2.0 ** 3.0
        assert unit_kpc.unit == "solMass / kpc^3"

        unit_arcsec = unit_kpc.convert(
            unit_length="arcsec", kpc_per_arcsec=2.0, unit_mass="solMass"
        )

        assert unit_arcsec == 2.0
        assert unit_arcsec.unit == "solMass / arcsec^3"

        with pytest.raises(exc.UnitsException):
            unit_arcsec.convert(unit_length="kpc", unit_mass="solMass")
            unit_kpc.convert(unit_length="arcsec", unit_mass="solMass")
            unit_arcsec.convert(unit_length="lol", unit_mass="solMass")


class MockDimensionsProfile(aast.dim.DimensionsProfile):
    def __init__(
        self,
        position: aast.dim.Position = None,
        param_float: float = None,
        length: aast.dim.Length = None,
        luminosity: aast.dim.Luminosity = None,
        mass: aast.dim.Mass = None,
        mass_over_luminosity: aast.dim.MassOverLuminosity = None,
    ):

        super(MockDimensionsProfile, self).__init__()

        self.position = position
        self.param_float = param_float
        self.luminosity = luminosity
        self.length = length
        self.mass = mass
        self.mass_over_luminosity = mass_over_luminosity


class TestDimensionsProfile:
    class TestUnitProperties:
        def test__extracts_length_correctly__raises_error_if_different_lengths_input(
            self
        ):
            profile = MockDimensionsProfile(
                position=(
                    aast.dim.Length(value=3.0, unit_length="arcsec"),
                    aast.dim.Length(value=3.0, unit_length="arcsec"),
                ),
                length=aast.dim.Length(3.0, "arcsec"),
            )

            assert profile.unit_length == "arcsec"

            profile = MockDimensionsProfile(
                position=(
                    aast.dim.Length(value=3.0, unit_length="kpc"),
                    aast.dim.Length(value=3.0, unit_length="kpc"),
                ),
                length=aast.dim.Length(3.0, "kpc"),
            )

            assert profile.unit_length == "kpc"

            with pytest.raises(exc.UnitsException):
                profile = MockDimensionsProfile(
                    position=(
                        aast.dim.Length(value=3.0, unit_length="kpc"),
                        aast.dim.Length(value=3.0, unit_length="kpc"),
                    ),
                    length=aast.dim.Length(3.0, "arcsec"),
                )

                profile.unit_length

        def test__extracts_luminosity_correctly__raises_error_if_different_luminosities(
            self
        ):
            profile = MockDimensionsProfile(
                luminosity=aast.dim.Luminosity(3.0, "eps"),
                mass_over_luminosity=aast.dim.MassOverLuminosity(
                    value=1.0, unit_luminosity="eps"
                ),
            )

            assert profile.unit_luminosity == "eps"

            profile = MockDimensionsProfile(
                luminosity=aast.dim.Luminosity(3.0, "counts"),
                mass_over_luminosity=aast.dim.MassOverLuminosity(
                    value=1.0, unit_luminosity="counts"
                ),
            )

            assert profile.unit_luminosity == "counts"

            with pytest.raises(exc.UnitsException):
                profile = MockDimensionsProfile(
                    luminosity=aast.dim.Luminosity(3.0, "eps"),
                    mass_over_luminosity=aast.dim.MassOverLuminosity(
                        value=1.0, unit_luminosity="counts"
                    ),
                )

                profile.unit_luminosity

        def test__extracts_mass_correctly__raises_error_if_different_mass(self):
            profile = MockDimensionsProfile(
                mass=aast.dim.Mass(3.0, "angular"),
                mass_over_luminosity=aast.dim.MassOverLuminosity(
                    value=1.0, unit_mass="angular"
                ),
            )

            assert profile.unit_mass == "angular"

            profile = MockDimensionsProfile(
                mass=aast.dim.Mass(3.0, "solMass"),
                mass_over_luminosity=aast.dim.MassOverLuminosity(
                    value=1.0, unit_mass="solMass"
                ),
            )

            assert profile.unit_mass == "solMass"

            with pytest.raises(exc.UnitsException):
                profile = MockDimensionsProfile(
                    mass=aast.dim.Mass(3.0, "angular"),
                    mass_over_luminosity=aast.dim.MassOverLuminosity(
                        value=1.0, unit_mass="solMass"
                    ),
                )

                profile.unit_mass

    class TestUnitConversions:
        def test__arcsec_to_kpc_conversions_of_length__float_and_tuple_length__conversion_converts_values(
            self
        ):

            profile_arcsec = MockDimensionsProfile(
                position=(
                    aast.dim.Length(1.0, "arcsec"),
                    aast.dim.Length(2.0, "arcsec"),
                ),
                param_float=2.0,
                length=aast.dim.Length(value=3.0, unit_length="arcsec"),
                luminosity=aast.dim.Luminosity(value=4.0, unit_luminosity="eps"),
                mass=aast.dim.Mass(value=5.0, unit_mass="angular"),
                mass_over_luminosity=aast.dim.MassOverLuminosity(
                    value=6.0, unit_luminosity="eps", unit_mass="angular"
                ),
            )

            assert profile_arcsec.position == (1.0, 2.0)
            assert profile_arcsec.position[0].unit_length == "arcsec"
            assert profile_arcsec.position[1].unit_length == "arcsec"
            assert profile_arcsec.param_float == 2.0
            assert profile_arcsec.length == 3.0
            assert profile_arcsec.length.unit_length == "arcsec"
            assert profile_arcsec.luminosity == 4.0
            assert profile_arcsec.luminosity.unit_luminosity == "eps"
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == "angular"
            assert profile_arcsec.mass_over_luminosity == 6.0
            assert profile_arcsec.mass_over_luminosity.unit == "angular / eps"

            profile_arcsec = profile_arcsec.new_object_with_units_converted(
                unit_length="arcsec"
            )

            assert profile_arcsec.position == (1.0, 2.0)
            assert profile_arcsec.position[0].unit == "arcsec"
            assert profile_arcsec.position[1].unit == "arcsec"
            assert profile_arcsec.param_float == 2.0
            assert profile_arcsec.length == 3.0
            assert profile_arcsec.length.unit == "arcsec"
            assert profile_arcsec.luminosity == 4.0
            assert profile_arcsec.luminosity.unit == "eps"
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == "angular"
            assert profile_arcsec.mass_over_luminosity == 6.0
            assert profile_arcsec.mass_over_luminosity.unit == "angular / eps"

            profile_kpc = profile_arcsec.new_object_with_units_converted(
                unit_length="kpc", kpc_per_arcsec=2.0
            )

            assert profile_kpc.position == (2.0, 4.0)
            assert profile_kpc.position[0].unit == "kpc"
            assert profile_kpc.position[1].unit == "kpc"
            assert profile_kpc.param_float == 2.0
            assert profile_kpc.length == 6.0
            assert profile_kpc.length.unit == "kpc"
            assert profile_kpc.luminosity == 4.0
            assert profile_kpc.luminosity.unit == "eps"
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == "angular"
            assert profile_kpc.mass_over_luminosity == 6.0
            assert profile_kpc.mass_over_luminosity.unit == "angular / eps"

            profile_kpc = profile_kpc.new_object_with_units_converted(unit_length="kpc")

            assert profile_kpc.position == (2.0, 4.0)
            assert profile_kpc.position[0].unit == "kpc"
            assert profile_kpc.position[1].unit == "kpc"
            assert profile_kpc.param_float == 2.0
            assert profile_kpc.length == 6.0
            assert profile_kpc.length.unit == "kpc"
            assert profile_kpc.luminosity == 4.0
            assert profile_kpc.luminosity.unit == "eps"
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == "angular"
            assert profile_kpc.mass_over_luminosity == 6.0
            assert profile_kpc.mass_over_luminosity.unit == "angular / eps"

            profile_arcsec = profile_kpc.new_object_with_units_converted(
                unit_length="arcsec", kpc_per_arcsec=2.0
            )

            assert profile_arcsec.position == (1.0, 2.0)
            assert profile_arcsec.position[0].unit == "arcsec"
            assert profile_arcsec.position[1].unit == "arcsec"
            assert profile_arcsec.param_float == 2.0
            assert profile_arcsec.length == 3.0
            assert profile_arcsec.length.unit == "arcsec"
            assert profile_arcsec.luminosity == 4.0
            assert profile_arcsec.luminosity.unit == "eps"
            assert profile_arcsec.mass == 5.0
            assert profile_arcsec.mass.unit_mass == "angular"
            assert profile_arcsec.mass_over_luminosity == 6.0
            assert profile_arcsec.mass_over_luminosity.unit == "angular / eps"

        def test__conversion_requires_kpc_per_arcsec_but_does_not_supply_it_raises_error(
            self
        ):

            profile_arcsec = MockDimensionsProfile(
                position=(
                    aast.dim.Length(1.0, "arcsec"),
                    aast.dim.Length(2.0, "arcsec"),
                )
            )

            with pytest.raises(exc.UnitsException):
                profile_arcsec.new_object_with_units_converted(unit_length="kpc")

            profile_kpc = profile_arcsec.new_object_with_units_converted(
                unit_length="kpc", kpc_per_arcsec=2.0
            )

            with pytest.raises(exc.UnitsException):
                profile_kpc.new_object_with_units_converted(unit_length="arcsec")

        def test__eps_to_counts_conversions_of_luminosity__conversions_convert_values(
            self
        ):

            profile_eps = MockDimensionsProfile(
                position=(
                    aast.dim.Length(1.0, "arcsec"),
                    aast.dim.Length(2.0, "arcsec"),
                ),
                param_float=2.0,
                length=aast.dim.Length(value=3.0, unit_length="arcsec"),
                luminosity=aast.dim.Luminosity(value=4.0, unit_luminosity="eps"),
                mass=aast.dim.Mass(value=5.0, unit_mass="angular"),
                mass_over_luminosity=aast.dim.MassOverLuminosity(
                    value=6.0, unit_luminosity="eps", unit_mass="angular"
                ),
            )

            assert profile_eps.position == (1.0, 2.0)
            assert profile_eps.position[0].unit_length == "arcsec"
            assert profile_eps.position[1].unit_length == "arcsec"
            assert profile_eps.param_float == 2.0
            assert profile_eps.length == 3.0
            assert profile_eps.length.unit_length == "arcsec"
            assert profile_eps.luminosity == 4.0
            assert profile_eps.luminosity.unit_luminosity == "eps"
            assert profile_eps.mass == 5.0
            assert profile_eps.mass.unit_mass == "angular"
            assert profile_eps.mass_over_luminosity == 6.0
            assert profile_eps.mass_over_luminosity.unit == "angular / eps"

            profile_eps = profile_eps.new_object_with_units_converted(
                unit_luminosity="eps"
            )

            assert profile_eps.position == (1.0, 2.0)
            assert profile_eps.position[0].unit_length == "arcsec"
            assert profile_eps.position[1].unit_length == "arcsec"
            assert profile_eps.param_float == 2.0
            assert profile_eps.length == 3.0
            assert profile_eps.length.unit_length == "arcsec"
            assert profile_eps.luminosity == 4.0
            assert profile_eps.luminosity.unit_luminosity == "eps"
            assert profile_eps.mass == 5.0
            assert profile_eps.mass.unit_mass == "angular"
            assert profile_eps.mass_over_luminosity == 6.0
            assert profile_eps.mass_over_luminosity.unit == "angular / eps"

            profile_counts = profile_eps.new_object_with_units_converted(
                unit_luminosity="counts", exposure_time=10.0
            )

            assert profile_counts.position == (1.0, 2.0)
            assert profile_counts.position[0].unit_length == "arcsec"
            assert profile_counts.position[1].unit_length == "arcsec"
            assert profile_counts.param_float == 2.0
            assert profile_counts.length == 3.0
            assert profile_counts.length.unit_length == "arcsec"
            assert profile_counts.luminosity == 40.0
            assert profile_counts.luminosity.unit_luminosity == "counts"
            assert profile_counts.mass == 5.0
            assert profile_counts.mass.unit_mass == "angular"
            assert profile_counts.mass_over_luminosity == pytest.approx(0.6, 1.0e-4)
            assert profile_counts.mass_over_luminosity.unit == "angular / counts"

            profile_counts = profile_counts.new_object_with_units_converted(
                unit_luminosity="counts"
            )

            assert profile_counts.position == (1.0, 2.0)
            assert profile_counts.position[0].unit_length == "arcsec"
            assert profile_counts.position[1].unit_length == "arcsec"
            assert profile_counts.param_float == 2.0
            assert profile_counts.length == 3.0
            assert profile_counts.length.unit_length == "arcsec"
            assert profile_counts.luminosity == 40.0
            assert profile_counts.luminosity.unit_luminosity == "counts"
            assert profile_counts.mass == 5.0
            assert profile_counts.mass.unit_mass == "angular"
            assert profile_counts.mass_over_luminosity == pytest.approx(0.6, 1.0e-4)
            assert profile_counts.mass_over_luminosity.unit == "angular / counts"

            profile_eps = profile_counts.new_object_with_units_converted(
                unit_luminosity="eps", exposure_time=10.0
            )

            assert profile_eps.position == (1.0, 2.0)
            assert profile_eps.position[0].unit_length == "arcsec"
            assert profile_eps.position[1].unit_length == "arcsec"
            assert profile_eps.param_float == 2.0
            assert profile_eps.length == 3.0
            assert profile_eps.length.unit_length == "arcsec"
            assert profile_eps.luminosity == 4.0
            assert profile_eps.luminosity.unit_luminosity == "eps"
            assert profile_eps.mass == 5.0
            assert profile_eps.mass.unit_mass == "angular"
            assert profile_eps.mass_over_luminosity == pytest.approx(6.0, 1.0e-4)
            assert profile_eps.mass_over_luminosity.unit == "angular / eps"

        def test__luminosity_conversion_requires_exposure_time_but_does_not_supply_it_raises_error(
            self
        ):

            profile_eps = MockDimensionsProfile(
                position=(
                    aast.dim.Length(1.0, "arcsec"),
                    aast.dim.Length(2.0, "arcsec"),
                ),
                param_float=2.0,
                length=aast.dim.Length(value=3.0, unit_length="arcsec"),
                luminosity=aast.dim.Luminosity(value=4.0, unit_luminosity="eps"),
                mass=aast.dim.Mass(value=5.0, unit_mass="angular"),
                mass_over_luminosity=aast.dim.MassOverLuminosity(
                    value=6.0, unit_luminosity="eps", unit_mass="angular"
                ),
            )

            with pytest.raises(exc.UnitsException):
                profile_eps.new_object_with_units_converted(unit_luminosity="counts")

            profile_counts = profile_eps.new_object_with_units_converted(
                unit_luminosity="counts", exposure_time=10.0
            )

            with pytest.raises(exc.UnitsException):
                profile_counts.new_object_with_units_converted(unit_luminosity="eps")

        def test__angular_to_solMass_conversions_of_mass__conversions_convert_values(
            self
        ):

            profile_angular = MockDimensionsProfile(
                position=(
                    aast.dim.Length(1.0, "arcsec"),
                    aast.dim.Length(2.0, "arcsec"),
                ),
                param_float=2.0,
                length=aast.dim.Length(value=3.0, unit_length="arcsec"),
                luminosity=aast.dim.Luminosity(value=4.0, unit_luminosity="eps"),
                mass=aast.dim.Mass(value=5.0, unit_mass="angular"),
                mass_over_luminosity=aast.dim.MassOverLuminosity(
                    value=6.0, unit_luminosity="eps", unit_mass="angular"
                ),
            )

            assert profile_angular.position == (1.0, 2.0)
            assert profile_angular.position[0].unit_length == "arcsec"
            assert profile_angular.position[1].unit_length == "arcsec"
            assert profile_angular.param_float == 2.0
            assert profile_angular.length == 3.0
            assert profile_angular.length.unit_length == "arcsec"
            assert profile_angular.luminosity == 4.0
            assert profile_angular.luminosity.unit_luminosity == "eps"
            assert profile_angular.mass == 5.0
            assert profile_angular.mass.unit_mass == "angular"
            assert profile_angular.mass_over_luminosity == 6.0
            assert profile_angular.mass_over_luminosity.unit == "angular / eps"

            profile_angular = profile_angular.new_object_with_units_converted(
                unit_mass="angular"
            )

            assert profile_angular.position == (1.0, 2.0)
            assert profile_angular.position[0].unit_length == "arcsec"
            assert profile_angular.position[1].unit_length == "arcsec"
            assert profile_angular.param_float == 2.0
            assert profile_angular.length == 3.0
            assert profile_angular.length.unit_length == "arcsec"
            assert profile_angular.luminosity == 4.0
            assert profile_angular.luminosity.unit_luminosity == "eps"
            assert profile_angular.mass == 5.0
            assert profile_angular.mass.unit_mass == "angular"
            assert profile_angular.mass_over_luminosity == 6.0
            assert profile_angular.mass_over_luminosity.unit == "angular / eps"

            profile_solMass = profile_angular.new_object_with_units_converted(
                unit_mass="solMass", critical_surface_density=10.0
            )

            assert profile_solMass.position == (1.0, 2.0)
            assert profile_solMass.position[0].unit_length == "arcsec"
            assert profile_solMass.position[1].unit_length == "arcsec"
            assert profile_solMass.param_float == 2.0
            assert profile_solMass.length == 3.0
            assert profile_solMass.length.unit_length == "arcsec"
            assert profile_solMass.luminosity == 4.0
            assert profile_solMass.luminosity.unit_luminosity == "eps"
            assert profile_solMass.mass == 50.0
            assert profile_solMass.mass.unit_mass == "solMass"
            assert profile_solMass.mass_over_luminosity == pytest.approx(60.0, 1.0e-4)
            assert profile_solMass.mass_over_luminosity.unit == "solMass / eps"

            profile_solMass = profile_solMass.new_object_with_units_converted(
                unit_mass="solMass"
            )

            assert profile_solMass.position == (1.0, 2.0)
            assert profile_solMass.position[0].unit_length == "arcsec"
            assert profile_solMass.position[1].unit_length == "arcsec"
            assert profile_solMass.param_float == 2.0
            assert profile_solMass.length == 3.0
            assert profile_solMass.length.unit_length == "arcsec"
            assert profile_solMass.luminosity == 4.0
            assert profile_solMass.luminosity.unit_luminosity == "eps"
            assert profile_solMass.mass == 50.0
            assert profile_solMass.mass.unit_mass == "solMass"
            assert profile_solMass.mass_over_luminosity == pytest.approx(60.0, 1.0e-4)
            assert profile_solMass.mass_over_luminosity.unit == "solMass / eps"

            profile_angular = profile_solMass.new_object_with_units_converted(
                unit_mass="angular", critical_surface_density=10.0
            )

            assert profile_angular.position == (1.0, 2.0)
            assert profile_angular.position[0].unit_length == "arcsec"
            assert profile_angular.position[1].unit_length == "arcsec"
            assert profile_angular.param_float == 2.0
            assert profile_angular.length == 3.0
            assert profile_angular.length.unit_length == "arcsec"
            assert profile_angular.luminosity == 4.0
            assert profile_angular.luminosity.unit_luminosity == "eps"
            assert profile_angular.mass == 5.0
            assert profile_angular.mass.unit_mass == "angular"
            assert profile_angular.mass_over_luminosity == pytest.approx(6.0, 1.0e-4)
            assert profile_angular.mass_over_luminosity.unit == "angular / eps"

        def test__mass_conversion_requires_critical_surface_density_but_does_not_supply_it_raises_error(
            self
        ):

            profile_angular = MockDimensionsProfile(
                position=(
                    aast.dim.Length(1.0, "arcsec"),
                    aast.dim.Length(2.0, "arcsec"),
                ),
                param_float=2.0,
                length=aast.dim.Length(value=3.0, unit_length="arcsec"),
                luminosity=aast.dim.Luminosity(value=4.0, unit_luminosity="eps"),
                mass=aast.dim.Mass(value=5.0, unit_mass="angular"),
                mass_over_luminosity=aast.dim.MassOverLuminosity(
                    value=6.0, unit_luminosity="eps", unit_mass="angular"
                ),
            )

            with pytest.raises(exc.UnitsException):
                profile_angular.new_object_with_units_converted(unit_mass="solMass")

            profile_solMass = profile_angular.new_object_with_units_converted(
                unit_mass="solMass", critical_surface_density=10.0
            )

            with pytest.raises(exc.UnitsException):
                profile_solMass.new_object_with_units_converted(unit_mass="angular")



import autofit as af
import autoarray as aa
import autoastro as aast
import math
import numpy as np
import pytest
import os

from test_autoastro.mock import mock_cosmology


@pytest.fixture(autouse=True)
def reset_config():
    """
    Use configuration from the default path. You may want to change this to set a specific path.
    """
    af.conf.instance = af.conf.default


def mass_within_radius_of_profile_from_grid_calculation(radius, profile):

    mass_total = 0.0

    xs = np.linspace(-radius * 1.5, radius * 1.5, 40)
    ys = np.linspace(-radius * 1.5, radius * 1.5, 40)

    edge = xs[1] - xs[0]
    area = edge ** 2

    for x in xs:
        for y in ys:

            eta = profile.grid_to_elliptical_radii(
                grid=aa.grid_irregular.manual_1d([[x, y]])
            )

            if eta < radius:
                mass_total += profile.convergence_func(eta) * area

    return mass_total


# class TestMassWithinCircle:
#     def test__mass_in_angular_units__singular_isothermal_sphere__compare_to_analytic(
#         self
#     ):
#
#         sis = aast.mp.SphericalIsothermal(einstein_radius=2.0)
#
#         radius = aast.dim.Length(2.0, "arcsec")
#
#         mass = sis.mass_within_circle_in_units(
#             radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
#         )
#         assert math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)
#
#         sis = aast.mp.SphericalIsothermal(einstein_radius=4.0)
#
#         radius = aast.dim.Length(4.0, "arcsec")
#
#         mass = sis.mass_within_circle_in_units(
#             radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
#         )
#         assert math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)
#
#     def test__mass_in_angular_units__singular_isothermal__compare_to_grid(self):
#
#         sis = aast.mp.SphericalIsothermal(einstein_radius=2.0)
#
#         radius = aast.dim.Length(1.0, "arcsec")
#
#         mass_grid = mass_within_radius_of_profile_from_grid_calculation(
#             radius=radius, profile=sis
#         )
#
#         mass = sis.mass_within_circle_in_units(
#             radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
#         )
#
#         assert mass_grid == pytest.approx(mass, 0.02)
#
#     def test__radius_units_conversions__mass_profile_updates_units_and_computes_correct_mass(
#         self
#     ):
#
#         cosmology = mock_cosmology.MockCosmology(kpc_per_arcsec=2.0)
#
#         # arcsec -> arcsec
#
#         sis_arcsec = aast.mp.SphericalIsothermal(
#             centre=(aast.dim.Length(0.0, "arcsec"), aast.dim.Length(0.0, "arcsec")),
#             einstein_radius=aast.dim.Length(2.0, "arcsec"),
#         )
#
#         radius = aast.dim.Length(2.0, "arcsec")
#         mass = sis_arcsec.mass_within_circle_in_units(
#             radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
#         )
#         assert math.pi * sis_arcsec.einstein_radius * radius == pytest.approx(
#             mass, 1e-3
#         )
#
#         # arcsec -> kpc
#
#         radius = aast.dim.Length(2.0, "kpc")
#         mass = sis_arcsec.mass_within_circle_in_units(
#             radius=radius,
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="angular",
#             cosmology=cosmology,
#         )
#         assert math.pi * sis_arcsec.einstein_radius * 1.0 == pytest.approx(mass, 1e-3)
#
#         # 2.0 arcsec = 4.0 kpc, same masses.
#
#         radius = aast.dim.Length(2.0, "arcsec")
#         mass_arcsec = sis_arcsec.mass_within_circle_in_units(
#             radius=radius,
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="angular",
#             cosmology=cosmology,
#         )
#         radius = aast.dim.Length(4.0, "kpc")
#         mass_kpc = sis_arcsec.mass_within_circle_in_units(
#             radius=radius,
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="angular",
#             cosmology=cosmology,
#         )
#         assert mass_arcsec == mass_kpc
#
#         # kpc -> kpc
#
#         sis_kpc = aast.mp.SphericalIsothermal(
#             centre=(aast.dim.Length(0.0, "kpc"), aast.dim.Length(0.0, "kpc")),
#             einstein_radius=aast.dim.Length(2.0, "kpc"),
#         )
#
#         sis_kpc.core_radius = aast.dim.Length(value=0.0, unit_length="kpc")
#
#         radius = aast.dim.Length(2.0, "kpc")
#         mass = sis_kpc.mass_within_circle_in_units(
#             radius=radius,
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="angular",
#             cosmology=cosmology,
#         )
#         assert math.pi * sis_kpc.einstein_radius * radius == pytest.approx(mass, 1e-3)
#
#         # kpc -> arcsec
#
#         radius = aast.dim.Length(2.0, "arcsec")
#         mass = sis_kpc.mass_within_circle_in_units(
#             radius=radius,
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="angular",
#             cosmology=cosmology,
#         )
#         assert 2.0 * math.pi * sis_kpc.einstein_radius * radius == pytest.approx(
#             mass, 1e-3
#         )
#
#         # 2.0 arcsec = 4.0 kpc, same masses.
#
#         radius = aast.dim.Length(2.0, "arcsec")
#         mass_arcsec = sis_kpc.mass_within_circle_in_units(
#             radius=radius,
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="angular",
#             cosmology=cosmology,
#         )
#         radius = aast.dim.Length(4.0, "kpc")
#         mass_kpc = sis_kpc.mass_within_circle_in_units(
#             radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
#         )
#         assert mass_arcsec == mass_kpc
#
#     def test__mass_units_conversions__multiplies_by_critical_surface_density_factor(
#         self
#     ):
#
#         cosmology = mock_cosmology.MockCosmology(critical_surface_density=2.0)
#
#         sis = aast.mp.SphericalIsothermal(einstein_radius=2.0)
#         radius = aast.dim.Length(2.0, "arcsec")
#
#         mass = sis.mass_within_circle_in_units(
#             radius=radius,
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="angular",
#             cosmology=cosmology,
#         )
#         assert math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)
#
#         mass = sis.mass_within_circle_in_units(
#             radius=radius,
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="solMass",
#             cosmology=cosmology,
#         )
#         assert 2.0 * math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)
#
#         mass = sis.mass_within_circle_in_units(
#             radius=radius,
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="solMass",
#             cosmology=cosmology,
#         )
#         assert 2.0 * math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)
#
#
# class TestDensityBetweenAnnuli:
#     def test__circular_annuli__sis__analyic_density_agrees(self):
#
#         cosmology = mock_cosmology.MockCosmology(
#             kpc_per_arcsec=2.0, critical_surface_density=2.0
#         )
#
#         einstein_radius = 1.0
#         sis_arcsec = aast.mp.SphericalIsothermal(
#             centre=(0.0, 0.0), einstein_radius=einstein_radius
#         )
#
#         inner_annuli_radius = aast.dim.Length(2.0, "arcsec")
#         outer_annuli_radius = aast.dim.Length(3.0, "arcsec")
#
#         inner_mass = math.pi * einstein_radius * inner_annuli_radius
#         outer_mass = math.pi * einstein_radius * outer_annuli_radius
#
#         density_between_annuli = sis_arcsec.density_between_circular_annuli_in_angular_units(
#             inner_annuli_radius=inner_annuli_radius,
#             outer_annuli_radius=outer_annuli_radius,
#             unit_length="arcsec",
#             unit_mass="angular",
#             redshift_profile=0.5,
#             redshift_source=1.0,
#             cosmology=cosmology,
#         )
#
#         annuli_area = (np.pi * outer_annuli_radius ** 2.0) - (
#             np.pi * inner_annuli_radius ** 2.0
#         )
#
#         assert (outer_mass - inner_mass) / annuli_area == pytest.approx(
#             density_between_annuli, 1e-4
#         )
#
#         density_between_annuli = sis_arcsec.density_between_circular_annuli_in_angular_units(
#             inner_annuli_radius=inner_annuli_radius,
#             outer_annuli_radius=outer_annuli_radius,
#             unit_length="arcsec",
#             unit_mass="solMass",
#             redshift_profile=0.5,
#             redshift_source=1.0,
#             cosmology=cosmology,
#         )
#
#         annuli_area = (np.pi * outer_annuli_radius ** 2.0) - (
#             np.pi * inner_annuli_radius ** 2.0
#         )
#
#         assert (2.0 * outer_mass - 2.0 * inner_mass) / annuli_area == pytest.approx(
#             density_between_annuli, 1e-4
#         )
#
#         density_between_annuli = sis_arcsec.density_between_circular_annuli_in_angular_units(
#             inner_annuli_radius=inner_annuli_radius,
#             outer_annuli_radius=outer_annuli_radius,
#             unit_length="kpc",
#             unit_mass="angular",
#             redshift_profile=0.5,
#             redshift_source=1.0,
#             cosmology=cosmology,
#         )
#
#         inner_mass = math.pi * 2.0 * einstein_radius * inner_annuli_radius
#         outer_mass = math.pi * 2.0 * einstein_radius * outer_annuli_radius
#
#         annuli_area = (np.pi * 2.0 * outer_annuli_radius ** 2.0) - (
#             np.pi * 2.0 * inner_annuli_radius ** 2.0
#         )
#
#         assert (outer_mass - inner_mass) / annuli_area == pytest.approx(
#             density_between_annuli, 1e-4
#         )
#
#     def test__circular_annuli__nfw_profile__compare_to_manual_mass(self):
#
#         cosmology = mock_cosmology.MockCosmology(
#             kpc_per_arcsec=2.0, critical_surface_density=2.0
#         )
#
#         nfw = aast.mp.EllipticalNFW(
#             centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, kappa_s=1.0
#         )
#
#         inner_mass = nfw.mass_within_circle_in_units(
#             radius=aast.dim.Length(1.0),
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="angular",
#         )
#
#         outer_mass = nfw.mass_within_circle_in_units(
#             radius=aast.dim.Length(2.0),
#             redshift_object=0.5,
#             redshift_source=1.0,
#             unit_mass="angular",
#         )
#
#         density_between_annuli = nfw.density_between_circular_annuli_in_angular_units(
#             inner_annuli_radius=aast.dim.Length(1.0),
#             outer_annuli_radius=aast.dim.Length(2.0),
#             unit_length="arcsec",
#             unit_mass="angular",
#             redshift_profile=0.5,
#             redshift_source=1.0,
#             cosmology=cosmology,
#         )
#
#         annuli_area = (np.pi * 2.0 ** 2.0) - (np.pi * 1.0 ** 2.0)
#
#         assert (outer_mass - inner_mass) / annuli_area == pytest.approx(
#             density_between_annuli, 1e-4
#         )
#
#
# class TestLensingObject:
#     def test__mass_profiles__list__is_list_of_self(self):
#
#         sis = aast.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)
#
#         assert sis.mass_profiles == [sis]
#
#     def test__correct_einstein_mass_caclulated__means_all_innherited_methods_work(
#         self,
#     ):
#
#         sis = aast.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)
#
#         assert sis.einstein_mass_in_units(unit_mass="angular") == pytest.approx(
#             np.pi * 2.0 ** 2.0, 1.0e-2
#         )



import numpy as np
import pytest
from skimage import measure

import autofit as af
import autoarray as aa
from autoarray.structures import grids
import autoastro as aast
from autoastro import exc
from test_autoastro.mock import mock_cosmology

import os


@pytest.fixture(autouse=True)
def reset_config():
    """
    Use configuration from the default path. You may want to change this to set a specific path.
    """
    af.conf.instance = af.conf.default


class TestUnits:
    def test__light_profiles_conversions(self):

        profile_0 = aast.lp.EllipticalGaussian(
            centre=(
                aast.dim.Length(value=3.0, unit_length="arcsec"),
                aast.dim.Length(value=3.0, unit_length="arcsec"),
            ),
            intensity=aast.dim.Luminosity(value=2.0, unit_luminosity="eps"),
        )

        profile_1 = aast.lp.EllipticalGaussian(
            centre=(
                aast.dim.Length(value=4.0, unit_length="arcsec"),
                aast.dim.Length(value=4.0, unit_length="arcsec"),
            ),
            intensity=aast.dim.Luminosity(value=5.0, unit_luminosity="eps"),
        )

        galaxy = aast.Galaxy(light_0=profile_0, light_1=profile_1, redshift=1.0)

        assert galaxy.light_0.centre == (3.0, 3.0)
        assert galaxy.light_0.unit_length == "arcsec"
        assert galaxy.light_0.intensity == 2.0
        assert galaxy.light_0.intensity.unit_luminosity == "eps"
        assert galaxy.light_1.centre == (4.0, 4.0)
        assert galaxy.light_1.unit_length == "arcsec"
        assert galaxy.light_1.intensity == 5.0
        assert galaxy.light_1.intensity.unit_luminosity == "eps"

        galaxy = galaxy.new_object_with_units_converted(
            unit_length="kpc",
            kpc_per_arcsec=2.0,
            unit_luminosity="counts",
            exposure_time=0.5,
        )

        assert galaxy.light_0.centre == (6.0, 6.0)
        assert galaxy.light_0.unit_length == "kpc"
        assert galaxy.light_0.intensity == 1.0
        assert galaxy.light_0.intensity.unit_luminosity == "counts"
        assert galaxy.light_1.centre == (8.0, 8.0)
        assert galaxy.light_1.unit_length == "kpc"
        assert galaxy.light_1.intensity == 2.5
        assert galaxy.light_1.intensity.unit_luminosity == "counts"

    def test__mass_profiles_conversions(self):

        profile_0 = aast.mp.EllipticalSersic(
            centre=(
                aast.dim.Length(value=3.0, unit_length="arcsec"),
                aast.dim.Length(value=3.0, unit_length="arcsec"),
            ),
            intensity=aast.dim.Luminosity(value=2.0, unit_luminosity="eps"),
            mass_to_light_ratio=aast.dim.MassOverLuminosity(
                value=5.0, unit_mass="angular", unit_luminosity="eps"
            ),
        )

        profile_1 = aast.mp.EllipticalSersic(
            centre=(
                aast.dim.Length(value=4.0, unit_length="arcsec"),
                aast.dim.Length(value=4.0, unit_length="arcsec"),
            ),
            intensity=aast.dim.Luminosity(value=5.0, unit_luminosity="eps"),
            mass_to_light_ratio=aast.dim.MassOverLuminosity(
                value=10.0, unit_mass="angular", unit_luminosity="eps"
            ),
        )

        galaxy = aast.Galaxy(mass_0=profile_0, mass_1=profile_1, redshift=1.0)

        assert galaxy.mass_0.centre == (3.0, 3.0)
        assert galaxy.mass_0.unit_length == "arcsec"
        assert galaxy.mass_0.intensity == 2.0
        assert galaxy.mass_0.intensity.unit_luminosity == "eps"
        assert galaxy.mass_0.mass_to_light_ratio == 5.0
        assert galaxy.mass_0.mass_to_light_ratio.unit_mass == "angular"
        assert galaxy.mass_1.centre == (4.0, 4.0)
        assert galaxy.mass_1.unit_length == "arcsec"
        assert galaxy.mass_1.intensity == 5.0
        assert galaxy.mass_1.intensity.unit_luminosity == "eps"
        assert galaxy.mass_1.mass_to_light_ratio == 10.0
        assert galaxy.mass_1.mass_to_light_ratio.unit_mass == "angular"

        galaxy = galaxy.new_object_with_units_converted(
            unit_length="kpc",
            kpc_per_arcsec=2.0,
            unit_luminosity="counts",
            exposure_time=0.5,
            unit_mass="solMass",
            critical_surface_density=3.0,
        )

        assert galaxy.mass_0.centre == (6.0, 6.0)
        assert galaxy.mass_0.unit_length == "kpc"
        assert galaxy.mass_0.intensity == 1.0
        assert galaxy.mass_0.intensity.unit_luminosity == "counts"
        assert galaxy.mass_0.mass_to_light_ratio == 30.0
        assert galaxy.mass_0.mass_to_light_ratio.unit_mass == "solMass"
        assert galaxy.mass_1.centre == (8.0, 8.0)
        assert galaxy.mass_1.unit_length == "kpc"
        assert galaxy.mass_1.intensity == 2.5
        assert galaxy.mass_1.intensity.unit_luminosity == "counts"
        assert galaxy.mass_1.mass_to_light_ratio == 60.0
        assert galaxy.mass_1.mass_to_light_ratio.unit_mass == "solMass"

    def test__galaxy_keeps_attributes(self):

        profile_0 = aast.lp.EllipticalGaussian(
            centre=(
                aast.dim.Length(value=3.0, unit_length="arcsec"),
                aast.dim.Length(value=3.0, unit_length="arcsec"),
            ),
            intensity=aast.dim.Luminosity(value=2.0, unit_luminosity="eps"),
        )

        galaxy = aast.Galaxy(
            light_0=profile_0, redshift=1.0, pixelization=1, regularization=2
        )

        assert galaxy.redshift == 1.0
        assert galaxy.pixelization == 1
        assert galaxy.regularization == 2

        galaxy = galaxy.new_object_with_units_converted(
            unit_length="kpc",
            kpc_per_arcsec=2.0,
            unit_luminosity="counts",
            exposure_time=0.5,
        )

        assert galaxy.redshift == 1.0
        assert galaxy.pixelization == 1
        assert galaxy.regularization == 2


class TestLightProfiles:
    class TestProfileImage:
        def test__no_light_profiles__profile_image_returned_as_0s_of_shape_grid(
            self, sub_grid_7x7
        ):
            galaxy = aast.Galaxy(redshift=0.5)

            profile_image = galaxy.profile_image_from_grid(grid=sub_grid_7x7)

            assert (profile_image == np.zeros(shape=sub_grid_7x7.sub_shape_1d)).all()

        def test__using_no_light_profiles__check_reshaping_decorator_of_returned_profile_image(
            self, sub_grid_7x7
        ):
            galaxy = aast.Galaxy(redshift=0.5)

            profile_image = galaxy.profile_image_from_grid(grid=sub_grid_7x7)

            assert (profile_image.in_2d_binned == np.zeros(shape=(7, 7))).all()

            profile_image = galaxy.profile_image_from_grid(grid=sub_grid_7x7)

            assert (profile_image == np.zeros(shape=sub_grid_7x7.sub_shape_1d)).all()

            profile_image = galaxy.profile_image_from_grid(grid=sub_grid_7x7)

            assert (
                profile_image.in_1d_binned
                == np.zeros(shape=sub_grid_7x7.sub_shape_1d // 4)
            ).all()

        def test__galaxies_with_x1_and_x2_light_profiles__profile_image_is_same_individual_profiles(
            self, lp_0, gal_x1_lp, lp_1, gal_x2_lp
        ):
            lp_profile_image = lp_0.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            gal_lp_profile_image = gal_x1_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            assert lp_profile_image == gal_lp_profile_image

            lp_profile_image = lp_0.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )
            lp_profile_image += lp_1.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            gal_profile_image = gal_x2_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            assert lp_profile_image == gal_profile_image

        def test__coordinates_in__coordinates_out(
            self, lp_0, gal_x1_lp, lp_1, gal_x2_lp
        ):
            lp_profile_image = lp_0.profile_image_from_grid(
                grid=aa.coordinates([[(1.05, -0.55)]])
            )

            gal_lp_profile_image = gal_x1_lp.profile_image_from_grid(
                grid=aa.coordinates([[(1.05, -0.55)]])
            )

            assert lp_profile_image[0][0] == gal_lp_profile_image[0][0]

        def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper_by_binning_sum_of_light_profile_values(
            self, sub_grid_7x7, gal_x2_lp
        ):

            lp_0_profile_image = gal_x2_lp.light_profile_0.profile_image_from_grid(
                grid=sub_grid_7x7
            )

            lp_1_profile_image = gal_x2_lp.light_profile_1.profile_image_from_grid(
                grid=sub_grid_7x7
            )

            lp_profile_image = lp_0_profile_image + lp_1_profile_image

            lp_profile_image_0 = (
                lp_profile_image[0]
                + lp_profile_image[1]
                + lp_profile_image[2]
                + lp_profile_image[3]
            ) / 4.0

            lp_profile_image_1 = (
                lp_profile_image[4]
                + lp_profile_image[5]
                + lp_profile_image[6]
                + lp_profile_image[7]
            ) / 4.0

            gal_profile_image = gal_x2_lp.profile_image_from_grid(grid=sub_grid_7x7)

            assert gal_profile_image.in_1d_binned[0] == lp_profile_image_0
            assert gal_profile_image.in_1d_binned[1] == lp_profile_image_1

    class TestLuminosityWithin:
        def test__two_profile_galaxy__is_sum_of_individual_profiles(
            self, lp_0, lp_1, gal_x1_lp, gal_x2_lp
        ):
            radius = aast.dim.Length(0.5, "arcsec")

            lp_luminosity = lp_0.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity="eps"
            )
            gal_luminosity = gal_x1_lp.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity="eps"
            )

            assert lp_luminosity == gal_luminosity

        def test__radius_unit_conversions__multiply_by_kpc_per_arcsec(
            self, lp_0, gal_x1_lp
        ):
            cosmology = mock_cosmology.MockCosmology(
                arcsec_per_kpc=0.5, kpc_per_arcsec=2.0
            )

            radius = aast.dim.Length(0.5, "arcsec")

            lp_luminosity_arcsec = lp_0.luminosity_within_circle_in_units(radius=radius)
            gal_luminosity_arcsec = gal_x1_lp.luminosity_within_circle_in_units(
                radius=radius
            )

            assert lp_luminosity_arcsec == gal_luminosity_arcsec

            radius = aast.dim.Length(0.5, "kpc")

            lp_luminosity_kpc = lp_0.luminosity_within_circle_in_units(
                radius=radius, redshift_object=0.5, cosmology=cosmology
            )
            gal_luminosity_kpc = gal_x1_lp.luminosity_within_circle_in_units(
                radius=radius, cosmology=cosmology
            )

            assert lp_luminosity_kpc == gal_luminosity_kpc

        def test__luminosity_unit_conversions__multiply_by_exposure_time(
            self, lp_0, gal_x1_lp
        ):

            radius = 3.0

            lp_luminosity_counts = lp_0.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity="counts", exposure_time=2.0
            )

            gal_luminosity_counts = gal_x1_lp.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity="counts", exposure_time=2.0
            )

            assert lp_luminosity_counts == gal_luminosity_counts

        def test__no_light_profile__returns_none(self):
            gal_no_lp = aast.Galaxy(redshift=0.5, mass=aast.mp.SphericalIsothermal())

            assert gal_no_lp.luminosity_within_circle_in_units(radius=1.0) == None

    class TestSymmetricProfiles:
        def test_1d_symmetry(self):
            lp_0 = aast.lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
            )

            lp_1 = aast.lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
                centre=(100, 0),
            )

            gal_x2_lp = aast.Galaxy(
                redshift=0.5, light_profile_0=lp_0, light_profile_1=lp_1
            )

            assert gal_x2_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[0.0, 0.0]])
            ) == gal_x2_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[100.0, 0.0]])
            )
            assert gal_x2_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 0.0]])
            ) == gal_x2_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[51.0, 0.0]])
            )

        def test_2d_symmetry(self):
            lp_0 = aast.lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
            )

            lp_1 = aast.lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
                centre=(100, 0),
            )

            lp_2 = aast.lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
                centre=(0, 100),
            )

            lp_3 = aast.lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
                centre=(100, 100),
            )

            gal_x4_lp = aast.Galaxy(
                redshift=0.5,
                light_profile_0=lp_0,
                light_profile_1=lp_1,
                light_profile_3=lp_2,
                light_profile_4=lp_3,
            )

            assert gal_x4_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_lp.profile_image_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 0.0]])
                ),
                1e-5,
            )

            assert gal_x4_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[0.0, 49.0]])
            ) == pytest.approx(
                gal_x4_lp.profile_image_from_grid(
                    grid=aa.grid_irregular.manual_1d([[0.0, 51.0]])
                ),
                1e-5,
            )

            assert gal_x4_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[100.0, 49.0]])
            ) == pytest.approx(
                gal_x4_lp.profile_image_from_grid(
                    grid=aa.grid_irregular.manual_1d([[100.0, 51.0]])
                ),
                1e-5,
            )

            assert gal_x4_lp.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 49.0]])
            ) == pytest.approx(
                gal_x4_lp.profile_image_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 51.0]])
                ),
                1e-5,
            )

    class TestBlurredProfileImages:
        def test__blurred_image_from_grid_and_psf(
            self, sub_grid_7x7, blurring_grid_7x7, psf_3x3, convolver_7x7
        ):
            light_profile_0 = aast.lp.EllipticalSersic(intensity=2.0)
            light_profile_1 = aast.lp.EllipticalSersic(intensity=3.0)

            galaxy = aast.Galaxy(
                light_profile_0=light_profile_0,
                light_profile_1=light_profile_1,
                redshift=0.5,
            )

            image = galaxy.profile_image_from_grid(grid=sub_grid_7x7)

            blurring_image = galaxy.profile_image_from_grid(grid=blurring_grid_7x7)

            blurred_image = convolver_7x7.convolved_image_from_image_and_blurring_image(
                image=image.in_1d_binned, blurring_image=blurring_image.in_1d_binned
            )

            light_profile_blurred_image = galaxy.blurred_profile_image_from_grid_and_psf(
                grid=sub_grid_7x7, blurring_grid=blurring_grid_7x7, psf=psf_3x3
            )

            assert blurred_image.in_1d == pytest.approx(
                light_profile_blurred_image.in_1d, 1.0e-4
            )

            assert blurred_image.in_2d == pytest.approx(
                light_profile_blurred_image.in_2d, 1.0e-4
            )

        def test__blurred_image_from_grid_and_convolver(
            self, sub_grid_7x7, blurring_grid_7x7, convolver_7x7
        ):
            light_profile_0 = aast.lp.EllipticalSersic(intensity=2.0)
            light_profile_1 = aast.lp.EllipticalSersic(intensity=3.0)

            galaxy = aast.Galaxy(
                light_profile_0=light_profile_0,
                light_profile_1=light_profile_1,
                redshift=0.5,
            )

            image = galaxy.profile_image_from_grid(grid=sub_grid_7x7)

            blurring_image = galaxy.profile_image_from_grid(grid=blurring_grid_7x7)

            blurred_image = convolver_7x7.convolved_image_from_image_and_blurring_image(
                image=image.in_1d_binned, blurring_image=blurring_image.in_1d_binned
            )

            light_profile_blurred_image = galaxy.blurred_profile_image_from_grid_and_convolver(
                grid=sub_grid_7x7,
                convolver=convolver_7x7,
                blurring_grid=blurring_grid_7x7,
            )

            assert blurred_image.in_1d == pytest.approx(
                light_profile_blurred_image.in_1d, 1.0e-4
            )

            assert blurred_image.in_2d == pytest.approx(
                light_profile_blurred_image.in_2d, 1.0e-4
            )

    class TestVisibilities:
        def test__visibilities_from_grid_and_transformer(
            self, sub_grid_7x7, transformer_7x7_7
        ):

            light_profile_0 = aast.lp.EllipticalSersic(intensity=2.0)
            light_profile_1 = aast.lp.EllipticalSersic(intensity=3.0)

            image = (
                light_profile_0.profile_image_from_grid(grid=sub_grid_7x7).in_1d_binned
                + light_profile_1.profile_image_from_grid(
                    grid=sub_grid_7x7
                ).in_1d_binned
            )

            visibilities = transformer_7x7_7.visibilities_from_image(image=image)

            galaxy = aast.Galaxy(
                light_profile_0=light_profile_0,
                light_profile_1=light_profile_1,
                redshift=0.5,
            )

            galaxy_visibilities = galaxy.profile_visibilities_from_grid_and_transformer(
                grid=sub_grid_7x7, transformer=transformer_7x7_7
            )

            assert visibilities == pytest.approx(galaxy_visibilities, 1.0e-4)

    class TestLightProfileGeometry:
        def test__extracts_centres_correctly(self):

            galaxy = aast.Galaxy(redshift=0.5)

            assert galaxy.light_profile_centres == []

            galaxy = aast.Galaxy(
                redshift=0.5, mp_0=aast.lp.EllipticalLightProfile(centre=(0.0, 1.0))
            )

            assert galaxy.light_profile_centres == [(0.0, 1.0)]

            galaxy = aast.Galaxy(
                redshift=0.5,
                mp_0=aast.lp.EllipticalLightProfile(centre=(0.0, 1.0)),
                mp_1=aast.lp.EllipticalLightProfile(centre=(2.0, 3.0)),
                mp_2=aast.lp.EllipticalLightProfile(centre=(4.0, 5.0)),
            )

            assert galaxy.light_profile_centres == [(0.0, 1.0), (2.0, 3.0), (4.0, 5.0)]

            galaxy = aast.Galaxy(
                redshift=0.5,
                mp_0=aast.lp.EllipticalLightProfile(centre=(0.0, 1.0)),
                mp_1=aast.lp.EllipticalLightProfile(centre=(2.0, 3.0)),
                lp_0=aast.mp.EllipticalMassProfile(centre=(-1.0, -2.0)),
                mp_2=aast.lp.EllipticalLightProfile(centre=(4.0, 5.0)),
            )

            assert galaxy.light_profile_centres == [(0.0, 1.0), (2.0, 3.0), (4.0, 5.0)]


class TestMassProfiles:
    class TestConvergence:
        def test__no_mass_profiles__convergence_returned_as_0s_of_shape_grid(
            self, sub_grid_7x7
        ):
            galaxy = aast.Galaxy(redshift=0.5)

            convergence = galaxy.convergence_from_grid(grid=sub_grid_7x7)

            assert (
                convergence.in_1d == np.zeros(shape=sub_grid_7x7.sub_shape_1d)
            ).all()

        def test__using_no_mass_profiles__check_reshaping_decorator_of_returned_convergence(
            self, sub_grid_7x7
        ):
            galaxy = aast.Galaxy(redshift=0.5)

            convergence = galaxy.convergence_from_grid(grid=sub_grid_7x7)

            assert (convergence.in_2d_binned == np.zeros(shape=(7, 7))).all()

            convergence = galaxy.convergence_from_grid(grid=sub_grid_7x7)

            assert (
                convergence.in_1d == np.zeros(shape=sub_grid_7x7.sub_shape_1d)
            ).all()

            convergence = galaxy.convergence_from_grid(grid=sub_grid_7x7)

            assert (
                convergence.in_1d_binned
                == np.zeros(shape=sub_grid_7x7.sub_shape_1d // 4)
            ).all()

        def test__galaxies_with_x1_and_x2_mass_profiles__convergence_is_same_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):
            mp_convergence = mp_0.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            gal_mp_convergence = gal_x1_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            assert mp_convergence == gal_mp_convergence

            mp_convergence = mp_0.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )
            mp_convergence += mp_1.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            gal_convergence = gal_x2_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            assert mp_convergence == gal_convergence

        def test__coordinates_in__coordinates_out(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):

            mp_convergence = mp_0.convergence_from_grid(
                grid=aa.coordinates([[(1.05, -0.55)]])
            )

            gal_mp_convergence = gal_x1_mp.convergence_from_grid(
                grid=aa.coordinates([[(1.05, -0.55)]])
            )

            assert mp_convergence == gal_mp_convergence

            assert type(gal_mp_convergence) == list
            assert mp_convergence[0][0] == gal_mp_convergence[0][0]

        def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper_by_binning_sum_of_mass_profile_values(
            self, sub_grid_7x7, gal_x2_mp
        ):
            mp_0_convergence = gal_x2_mp.mass_profile_0.convergence_from_grid(
                grid=sub_grid_7x7
            )

            mp_1_convergence = gal_x2_mp.mass_profile_1.convergence_from_grid(
                grid=sub_grid_7x7
            )

            mp_convergence = mp_0_convergence + mp_1_convergence

            mp_convergence_0 = (
                mp_convergence[0]
                + mp_convergence[1]
                + mp_convergence[2]
                + mp_convergence[3]
            ) / 4.0

            mp_convergence_1 = (
                mp_convergence[4]
                + mp_convergence[5]
                + mp_convergence[6]
                + mp_convergence[7]
            ) / 4.0

            gal_convergence = gal_x2_mp.convergence_from_grid(grid=sub_grid_7x7)

            assert gal_convergence.in_1d_binned[0] == mp_convergence_0
            assert gal_convergence.in_1d_binned[1] == mp_convergence_1

    class TestPotential:
        def test__no_mass_profiles__potential_returned_as_0s_of_shape_grid(
            self, sub_grid_7x7
        ):
            galaxy = aast.Galaxy(redshift=0.5)

            potential = galaxy.potential_from_grid(grid=sub_grid_7x7)

            assert (potential.in_1d == np.zeros(shape=sub_grid_7x7.sub_shape_1d)).all()

        def test__using_no_mass_profiles__check_reshaping_decorator_of_returned_potential(
            self, sub_grid_7x7
        ):
            galaxy = aast.Galaxy(redshift=0.5)

            potential = galaxy.potential_from_grid(grid=sub_grid_7x7)

            assert (potential.in_2d_binned == np.zeros(shape=(7, 7))).all()

            potential = galaxy.potential_from_grid(grid=sub_grid_7x7)

            assert (potential.in_1d == np.zeros(shape=sub_grid_7x7.sub_shape_1d)).all()

            potential = galaxy.potential_from_grid(grid=sub_grid_7x7)

            assert (
                potential.in_1d_binned == np.zeros(shape=sub_grid_7x7.sub_shape_1d // 4)
            ).all()

        def test__galaxies_with_x1_and_x2_mass_profiles__potential_is_same_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):
            mp_potential = mp_0.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            gal_mp_potential = gal_x1_mp.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            assert mp_potential == gal_mp_potential

            mp_potential = mp_0.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )
            mp_potential += mp_1.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            gal_potential = gal_x2_mp.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            assert mp_potential == gal_potential

        def test__coordinates_in__coordinates_out(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):

            mp_potential = mp_0.potential_from_grid(
                grid=aa.coordinates([[(1.05, -0.55)]])
            )

            gal_mp_potential = gal_x1_mp.potential_from_grid(
                grid=aa.coordinates([[(1.05, -0.55)]])
            )

            assert mp_potential == gal_mp_potential

            assert type(gal_mp_potential) == list
            assert mp_potential[0][0] == gal_mp_potential[0][0]

        def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper_by_binning_sum_of_mass_profile_values(
            self, sub_grid_7x7, gal_x2_mp
        ):
            mp_0_potential = gal_x2_mp.mass_profile_0.potential_from_grid(
                grid=sub_grid_7x7
            )

            mp_1_potential = gal_x2_mp.mass_profile_1.potential_from_grid(
                grid=sub_grid_7x7
            )

            mp_potential = mp_0_potential + mp_1_potential

            mp_potential_0 = (
                mp_potential[0] + mp_potential[1] + mp_potential[2] + mp_potential[3]
            ) / 4.0

            mp_potential_1 = (
                mp_potential[4] + mp_potential[5] + mp_potential[6] + mp_potential[7]
            ) / 4.0

            gal_potential = gal_x2_mp.potential_from_grid(grid=sub_grid_7x7)

            assert gal_potential.in_1d_binned[0] == mp_potential_0
            assert gal_potential.in_1d_binned[1] == mp_potential_1

    class TestDeflectionAngles:
        def test__no_mass_profiles__deflections_returned_as_0s_of_shape_grid(
            self, sub_grid_7x7
        ):
            galaxy = aast.Galaxy(redshift=0.5)

            deflections = galaxy.deflections_from_grid(grid=sub_grid_7x7)

            assert (
                deflections.in_1d == np.zeros(shape=(sub_grid_7x7.sub_shape_1d, 2))
            ).all()

        def test__using_no_mass_profiles__check_reshaping_decorator_of_returned_deflections(
            self, sub_grid_7x7
        ):
            galaxy = aast.Galaxy(redshift=0.5)

            deflections = galaxy.deflections_from_grid(grid=sub_grid_7x7)

            assert (deflections.in_2d_binned == np.zeros(shape=(7, 7, 2))).all()

            deflections = galaxy.deflections_from_grid(grid=sub_grid_7x7)

            assert (
                deflections.in_1d == np.zeros(shape=(sub_grid_7x7.sub_shape_1d, 2))
            ).all()

            deflections = galaxy.deflections_from_grid(grid=sub_grid_7x7)

            assert (
                deflections.in_1d_binned
                == np.zeros(shape=(sub_grid_7x7.sub_shape_1d // 4, 2))
            ).all()

        def test__galaxies_with_x1_and_x2_mass_profiles__deflections_is_same_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):
            mp_deflections = mp_0.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            gal_mp_deflections = gal_x1_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            assert (mp_deflections == gal_mp_deflections).all()

            mp_deflections = mp_0.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )
            mp_deflections += mp_1.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            gal_deflections = gal_x2_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.05, -0.55]])
            )

            assert (mp_deflections == gal_deflections).all()

        def test__coordinates_in__coordinates_out(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):

            mp_deflections = mp_0.deflections_from_grid(
                grid=aa.coordinates([[(1.05, -0.55)]])
            )

            gal_mp_deflections = gal_x1_mp.deflections_from_grid(
                grid=aa.coordinates([[(1.05, -0.55)]])
            )

            assert mp_deflections == gal_mp_deflections

            assert type(gal_mp_deflections) == grids.Coordinates
            assert mp_deflections[0][0][0] == gal_mp_deflections[0][0][0]
            assert mp_deflections[0][0][1] == gal_mp_deflections[0][0][1]

        def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper_by_binning_sum_of_mass_profile_values(
            self, sub_grid_7x7, gal_x2_mp
        ):
            mp_0_deflections = gal_x2_mp.mass_profile_0.deflections_from_grid(
                grid=sub_grid_7x7
            )

            mp_1_deflections = gal_x2_mp.mass_profile_1.deflections_from_grid(
                grid=sub_grid_7x7
            )

            mp_deflections = mp_0_deflections + mp_1_deflections

            mp_deflections_y_0 = (
                mp_deflections[0, 0]
                + mp_deflections[1, 0]
                + mp_deflections[2, 0]
                + mp_deflections[3, 0]
            ) / 4.0

            mp_deflections_y_1 = (
                mp_deflections[4, 0]
                + mp_deflections[5, 0]
                + mp_deflections[6, 0]
                + mp_deflections[7, 0]
            ) / 4.0

            gal_deflections = gal_x2_mp.deflections_from_grid(grid=sub_grid_7x7)

            assert gal_deflections.in_1d_binned[0, 0] == mp_deflections_y_0
            assert gal_deflections.in_1d_binned[1, 0] == mp_deflections_y_1

            mp_deflections_x_0 = (
                mp_deflections[0, 1]
                + mp_deflections[1, 1]
                + mp_deflections[2, 1]
                + mp_deflections[3, 1]
            ) / 4.0

            mp_deflections_x_1 = (
                mp_deflections[4, 1]
                + mp_deflections[5, 1]
                + mp_deflections[6, 1]
                + mp_deflections[7, 1]
            ) / 4.0

            gal_deflections = gal_x2_mp.deflections_from_grid(grid=sub_grid_7x7)

            assert gal_deflections.in_1d_binned[0, 1] == mp_deflections_x_0
            assert gal_deflections.in_1d_binned[1, 1] == mp_deflections_x_1

    class TestMassWithin:
        def test__two_profile_galaxy__is_sum_of_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):
            radius = aast.dim.Length(0.5, "arcsec")

            mp_mass = mp_0.mass_within_circle_in_units(
                radius=radius, unit_mass="angular"
            )

            gal_mass = gal_x1_mp.mass_within_circle_in_units(
                radius=radius, unit_mass="angular"
            )

            assert mp_mass == gal_mass

        def test__radius_unit_conversions__multiply_by_kpc_per_arcsec(
            self, mp_0, gal_x1_mp
        ):
            cosmology = mock_cosmology.MockCosmology(
                arcsec_per_kpc=0.5, kpc_per_arcsec=2.0, critical_surface_density=1.0
            )

            radius = aast.dim.Length(0.5, "arcsec")

            mp_mass_arcsec = mp_0.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_object=0.5,
                redshift_source=1.0,
                cosmology=cosmology,
            )

            gal_mass_arcsec = gal_x1_mp.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_source=1.0,
                cosmology=cosmology,
            )
            assert mp_mass_arcsec == gal_mass_arcsec

            radius = aast.dim.Length(0.5, "kpc")

            mp_mass_kpc = mp_0.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_object=0.5,
                redshift_source=1.0,
                cosmology=cosmology,
            )

            gal_mass_kpc = gal_x1_mp.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_source=1.0,
                cosmology=cosmology,
            )
            assert mp_mass_kpc == gal_mass_kpc

        def test__mass_unit_conversions__same_as_individual_profile(
            self, mp_0, gal_x1_mp
        ):
            cosmology = mock_cosmology.MockCosmology(
                arcsec_per_kpc=1.0, kpc_per_arcsec=1.0, critical_surface_density=2.0
            )

            radius = aast.dim.Length(0.5, "arcsec")

            mp_mass_sol = mp_0.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_object=0.5,
                redshift_source=1.0,
                cosmology=cosmology,
            )

            gal_mass_sol = gal_x1_mp.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_source=1.0,
                cosmology=cosmology,
            )
            assert mp_mass_sol == gal_mass_sol

        def test__no_mass_profile__returns_none(self):
            gal_no_mp = aast.Galaxy(redshift=0.5, light=aast.lp.SphericalSersic())

            with pytest.raises(exc.GalaxyException):
                gal_no_mp.mass_within_circle_in_units(radius=1.0)

    class TestSymmetricProfiles:
        def test_1d_symmetry(self):
            mp_0 = aast.mp.EllipticalIsothermal(
                axis_ratio=0.5, phi=45.0, einstein_radius=1.0
            )

            mp_1 = aast.mp.EllipticalIsothermal(
                centre=(100, 0), axis_ratio=0.5, phi=45.0, einstein_radius=1.0
            )

            gal_x4_mp = aast.Galaxy(
                redshift=0.5, mass_profile_0=mp_0, mass_profile_1=mp_1
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
            ) == gal_x4_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[99.0, 0.0]])
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 0.0]])
            ) == gal_x4_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[51.0, 0.0]])
            )

            assert gal_x4_mp.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(
                    grid=aa.grid_irregular.manual_1d([[99.0, 0.0]])
                ),
                1e-6,
            )

            assert gal_x4_mp.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 0.0]])
                ),
                1e-6,
            )

            assert gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[99.0, 0.0]])
                ),
                1e-6,
            )

            assert gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 0.0]])
                ),
                1e-6,
            )

        def test_2d_symmetry(self):
            mp_0 = aast.mp.SphericalIsothermal(einstein_radius=1.0)

            mp_1 = aast.mp.SphericalIsothermal(centre=(100, 0), einstein_radius=1.0)

            mp_2 = aast.mp.SphericalIsothermal(centre=(0, 100), einstein_radius=1.0)

            mp_3 = aast.mp.SphericalIsothermal(centre=(100, 100), einstein_radius=1.0)

            gal_x4_mp = aast.Galaxy(
                redshift=0.5,
                mass_profile_0=mp_0,
                mass_profile_1=mp_1,
                mass_profile_2=mp_2,
                mass_profile_3=mp_3,
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.convergence_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 0.0]])
                ),
                1e-5,
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[0.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.convergence_from_grid(
                    grid=aa.grid_irregular.manual_1d([[0.0, 51.0]])
                ),
                1e-5,
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[100.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.convergence_from_grid(
                    grid=aa.grid_irregular.manual_1d([[100.0, 51.0]])
                ),
                1e-5,
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.convergence_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 51.0]])
                ),
                1e-5,
            )

            assert gal_x4_mp.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 0.0]])
                ),
                1e-5,
            )

            assert gal_x4_mp.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[0.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(
                    grid=aa.grid_irregular.manual_1d([[0.0, 51.0]])
                ),
                1e-5,
            )

            assert gal_x4_mp.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[100.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(
                    grid=aa.grid_irregular.manual_1d([[100.0, 51.0]])
                ),
                1e-5,
            )

            assert gal_x4_mp.potential_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 51.0]])
                ),
                1e-5,
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 0.0]])
            )[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 0.0]])
                )[0, 0],
                1e-5,
            )

            assert 1.0 * gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[0.0, 49.0]])
            )[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[0.0, 51.0]])
                )[0, 0],
                1e-5,
            )

            assert 1.0 * gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[100.0, 49.0]])
            )[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[100.0, 51.0]])
                )[0, 0],
                1e-5,
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 49.0]])
            )[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 51.0]])
                )[0, 0],
                1e-5,
            )

            assert 1.0 * gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 0.0]])
            )[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 0.0]])
                )[0, 1],
                1e-5,
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[0.0, 49.0]])
            )[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[0.0, 51.0]])
                )[0, 1],
                1e-5,
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[100.0, 49.0]])
            )[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[100.0, 51.0]])
                )[0, 1],
                1e-5,
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(
                grid=aa.grid_irregular.manual_1d([[49.0, 49.0]])
            )[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(
                    grid=aa.grid_irregular.manual_1d([[51.0, 51.0]])
                )[0, 1],
                1e-5,
            )

    class TestMassProfileGeometry:
        def test__extracts_centres_correctly__ignores_mass_sheets(self):

            galaxy = aast.Galaxy(redshift=0.5)

            assert galaxy.mass_profile_centres == []

            galaxy = aast.Galaxy(
                redshift=0.5, mp_0=aast.mp.EllipticalMassProfile(centre=(0.0, 1.0))
            )

            assert galaxy.mass_profile_centres == [(0.0, 1.0)]

            galaxy = aast.Galaxy(
                redshift=0.5,
                mp_0=aast.mp.EllipticalMassProfile(centre=(0.0, 1.0)),
                mp_1=aast.mp.EllipticalMassProfile(centre=(2.0, 3.0)),
                mp_2=aast.mp.EllipticalMassProfile(centre=(4.0, 5.0)),
            )

            assert galaxy.mass_profile_centres == [(0.0, 1.0), (2.0, 3.0), (4.0, 5.0)]

            galaxy = aast.Galaxy(
                redshift=0.5,
                mp_0=aast.mp.EllipticalMassProfile(centre=(0.0, 1.0)),
                mp_1=aast.mp.EllipticalMassProfile(centre=(2.0, 3.0)),
                lp_0=aast.lp.EllipticalLightProfile(centre=(-1.0, -2.0)),
                mp_2=aast.mp.EllipticalMassProfile(centre=(4.0, 5.0)),
            )

            assert galaxy.mass_profile_centres == [(0.0, 1.0), (2.0, 3.0), (4.0, 5.0)]

            galaxy = aast.Galaxy(
                redshift=0.5,
                mp_0=aast.lp.EllipticalLightProfile(centre=(0.0, 1.0)),
                mp_1=aast.lp.EllipticalLightProfile(centre=(2.0, 3.0)),
                lp_0=aast.mp.EllipticalMassProfile(centre=(-1.0, -2.0)),
                mp_2=aast.lp.EllipticalLightProfile(centre=(4.0, 5.0)),
                sheet=aast.mp.MassSheet(centre=(10.0, 10.0)),
            )

            assert galaxy.light_profile_centres == [(0.0, 1.0), (2.0, 3.0), (4.0, 5.0)]

        def test__extracts_axis_ratio_correctly(self):

            galaxy = aast.Galaxy(redshift=0.5)

            assert galaxy.mass_profile_axis_ratios == []

            galaxy = aast.Galaxy(
                redshift=0.5, mp_0=aast.mp.EllipticalMassProfile(axis_ratio=0.9)
            )

            assert galaxy.mass_profile_axis_ratios == [0.9]

            galaxy = aast.Galaxy(
                redshift=0.5,
                mp_0=aast.mp.EllipticalMassProfile(axis_ratio=0.9),
                mp_1=aast.mp.EllipticalMassProfile(axis_ratio=0.8),
                mp_2=aast.mp.EllipticalMassProfile(axis_ratio=0.7),
            )

            assert galaxy.mass_profile_axis_ratios == [0.9, 0.8, 0.7]

            galaxy = aast.Galaxy(
                redshift=0.5,
                mp_0=aast.mp.EllipticalMassProfile(axis_ratio=0.9),
                mp_1=aast.mp.EllipticalMassProfile(axis_ratio=0.8),
                lp_0=aast.lp.EllipticalLightProfile(axis_ratio=0.1),
                mp_2=aast.mp.EllipticalMassProfile(axis_ratio=0.7),
            )

            assert galaxy.mass_profile_axis_ratios == [0.9, 0.8, 0.7]

        def test__extracts_phis_correctly(self):

            galaxy = aast.Galaxy(redshift=0.5)

            assert galaxy.mass_profile_phis == []

            galaxy = aast.Galaxy(
                redshift=0.5, mp_0=aast.mp.EllipticalMassProfile(phi=0.9)
            )

            assert galaxy.mass_profile_phis == [0.9]

            galaxy = aast.Galaxy(
                redshift=0.5,
                mp_0=aast.mp.EllipticalMassProfile(phi=0.9),
                mp_1=aast.mp.EllipticalMassProfile(phi=0.8),
                mp_2=aast.mp.EllipticalMassProfile(phi=0.7),
            )

            assert galaxy.mass_profile_phis == [0.9, 0.8, 0.7]

            galaxy = aast.Galaxy(
                redshift=0.5,
                mp_0=aast.mp.EllipticalMassProfile(phi=0.9),
                mp_1=aast.mp.EllipticalMassProfile(phi=0.8),
                lp_0=aast.lp.EllipticalLightProfile(phi=0.1),
                mp_2=aast.mp.EllipticalMassProfile(phi=0.7),
            )

            assert galaxy.mass_profile_phis == [0.9, 0.8, 0.7]

    class TestLensingObject:
        def test__correct_einstein_mass_caclulated_for_multiple_mass_profiles__means_all_innherited_methods_work(
            self,
        ):

            sis_0 = aast.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

            sis_1 = aast.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)

            galaxy = aast.Galaxy(
                mass_profile_0=sis_0, mass_profile_1=sis_1, redshift=0.5
            )

            assert galaxy.einstein_mass_in_units(unit_mass="angular") == pytest.approx(
                np.pi * 3.0 ** 2.0, 1.0e-2
            )


class TestMassAndLightProfiles:
    def test_single_profile(self, lmp_0):
        gal_x1_lmp = aast.Galaxy(redshift=0.5, profile=lmp_0)

        assert 1 == len(gal_x1_lmp.light_profiles)
        assert 1 == len(gal_x1_lmp.mass_profiles)

        assert gal_x1_lmp.mass_profiles[0] == lmp_0
        assert gal_x1_lmp.light_profiles[0] == lmp_0

    def test_multiple_profile(self, lmp_0, lp_0, mp_0):
        gal_multi_profiles = aast.Galaxy(
            redshift=0.5, profile=lmp_0, light=lp_0, sie=mp_0
        )

        assert 2 == len(gal_multi_profiles.light_profiles)
        assert 2 == len(gal_multi_profiles.mass_profiles)


class TestSummarizeInUnits:
    def test__galaxy_with_two_light_and_mass_profiles(self, lp_0, lp_1, mp_0, mp_1):

        test_path = "{}/../test_files/config/summary".format(
            os.path.dirname(os.path.realpath(__file__))
        )
        af.conf.instance = af.conf.Config(config_path=test_path)

        gal_summarize = aast.Galaxy(
            redshift=0.5,
            light_profile_0=lp_0,
            light_profile_1=lp_1,
            mass_profile_0=mp_0,
            mass_profile_1=mp_1,
        )

        summary_text = gal_summarize.summarize_in_units(
            radii=[aast.dim.Length(10.0), aast.dim.Length(500.0)],
            whitespace=50,
            unit_length="arcsec",
            unit_luminosity="eps",
            unit_mass="angular",
        )

        i = 0

        assert summary_text[i] == "Galaxy\n"
        i += 1
        assert (
            summary_text[i] == "redshift                                          0.50"
        )
        i += 1
        assert summary_text[i] == "\nGALAXY LIGHT\n\n"
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_10.00_arcsec                    1.8854e+02 eps"
        )
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_500.00_arcsec                   1.9573e+02 eps"
        )
        i += 1
        assert summary_text[i] == "\nLIGHT PROFILES:\n\n"
        i += 1
        assert summary_text[i] == "Light Profile = SphericalSersic\n"
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_10.00_arcsec                    6.2848e+01 eps"
        )
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_500.00_arcsec                   6.5243e+01 eps"
        )
        i += 1
        assert summary_text[i] == "\n"
        i += 1
        assert summary_text[i] == "Light Profile = SphericalSersic\n"
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_10.00_arcsec                    1.2570e+02 eps"
        )
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_500.00_arcsec                   1.3049e+02 eps"
        )
        i += 1
        assert summary_text[i] == "\n"
        i += 1
        assert summary_text[i] == "\nGALAXY MASS\n\n"
        i += 1
        assert (
            summary_text[i]
            == "einstein_radius                                   2.99 arcsec"
        )
        i += 1
        assert (
            summary_text[i]
            == "einstein_mass                                     2.8177e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_10.00_arcsec                          9.4248e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_500.00_arcsec                         4.7124e+03 angular"
        )
        i += 1
        assert summary_text[i] == "\nMASS PROFILES:\n\n"
        i += 1
        assert summary_text[i] == "Mass Profile = SphericalIsothermal\n"
        i += 1
        assert (
            summary_text[i]
            == "einstein_radius                                   1.00 arcsec"
        )
        i += 1
        assert (
            summary_text[i]
            == "einstein_mass                                     3.1308e+00 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_10.00_arcsec                          3.1416e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_500.00_arcsec                         1.5708e+03 angular"
        )
        i += 1
        assert summary_text[i] == "\n"
        i += 1
        assert summary_text[i] == "Mass Profile = SphericalIsothermal\n"
        i += 1
        assert (
            summary_text[i]
            == "einstein_radius                                   2.00 arcsec"
        )
        i += 1
        assert (
            summary_text[i]
            == "einstein_mass                                     1.2523e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_10.00_arcsec                          6.2832e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_500.00_arcsec                         3.1416e+03 angular"
        )
        i += 1
        assert summary_text[i] == "\n"
        i += 1


class TestHyperGalaxy:
    class TestContributionMaps:
        def test__model_image_all_1s__factor_is_0__contributions_all_1s(self):
            hyper_image = np.ones((3,))

            hyp = aast.HyperGalaxy(contribution_factor=0.0)
            contribution_map = hyp.contribution_map_from_hyper_images(
                hyper_model_image=hyper_image, hyper_galaxy_image=hyper_image
            )

            assert (contribution_map == np.ones((3,))).all()

        def test__different_values__factor_is_1__contributions_are_value_divided_by_factor_and_max(
            self
        ):
            hyper_image = np.array([0.5, 1.0, 1.5])

            hyp = aast.HyperGalaxy(contribution_factor=1.0)
            contribution_map = hyp.contribution_map_from_hyper_images(
                hyper_model_image=hyper_image, hyper_galaxy_image=hyper_image
            )

            assert (
                contribution_map
                == np.array([(0.5 / 1.5) / (1.5 / 2.5), (1.0 / 2.0) / (1.5 / 2.5), 1.0])
            ).all()

        def test__galaxy_contribution_map_property(self):

            hyper_image = np.ones((3,))

            hyp = aast.HyperGalaxy(contribution_factor=0.0)

            galaxy = aast.Galaxy(
                redshift=0.5,
                hyper_galaxy=hyp,
                hyper_galaxy_image=hyper_image,
                hyper_model_image=hyper_image,
            )

            contribution_map = hyp.contribution_map_from_hyper_images(
                hyper_model_image=hyper_image, hyper_galaxy_image=hyper_image
            )

            assert (contribution_map == galaxy.contribution_map).all()

    class TestHyperNoiseMap:
        def test__contribution_all_1s__noise_factor_2__noise_adds_double(self):
            noise_map = np.array([1.0, 2.0, 3.0])
            contribution_map = np.ones((3, 1))

            hyper_galaxy = aast.HyperGalaxy(
                contribution_factor=0.0, noise_factor=2.0, noise_power=1.0
            )

            hyper_noise_map = hyper_galaxy.hyper_noise_map_from_contribution_map(
                noise_map=noise_map, contribution_map=contribution_map
            )

            assert (hyper_noise_map == np.array([2.0, 4.0, 6.0])).all()

        def test__same_as_above_but_contributions_vary(self):
            noise_map = np.array([1.0, 2.0, 3.0])
            contribution_map = np.array([[0.0, 0.5, 1.0]])

            hyper_galaxy = aast.HyperGalaxy(
                contribution_factor=0.0, noise_factor=2.0, noise_power=1.0
            )

            hyper_noise_map = hyper_galaxy.hyper_noise_map_from_contribution_map(
                noise_map=noise_map, contribution_map=contribution_map
            )

            assert (hyper_noise_map == np.array([0.0, 2.0, 6.0])).all()

        def test__same_as_above_but_change_noise_scale_terms(self):
            noise_map = np.array([1.0, 2.0, 3.0])
            contribution_map = np.array([[0.0, 0.5, 1.0]])

            hyper_galaxy = aast.HyperGalaxy(
                contribution_factor=0.0, noise_factor=2.0, noise_power=2.0
            )

            hyper_noise_map = hyper_galaxy.hyper_noise_map_from_contribution_map(
                noise_map=noise_map, contribution_map=contribution_map
            )

            assert (hyper_noise_map == np.array([0.0, 2.0, 18.0])).all()


class TestBooleanProperties:
    def test_has_profile(self):
        assert aast.Galaxy(redshift=0.5).has_profile is False
        assert (
            aast.Galaxy(redshift=0.5, light_profile=aast.lp.LightProfile()).has_profile
            is True
        )
        assert (
            aast.Galaxy(redshift=0.5, mass_profile=aast.mp.MassProfile()).has_profile
            is True
        )

    def test_has_light_profile(self):
        assert aast.Galaxy(redshift=0.5).has_light_profile is False
        assert (
            aast.Galaxy(
                redshift=0.5, light_profile=aast.lp.LightProfile()
            ).has_light_profile
            is True
        )
        assert (
            aast.Galaxy(
                redshift=0.5, mass_profile=aast.mp.MassProfile()
            ).has_light_profile
            is False
        )

    def test_has_mass_profile(self):
        assert aast.Galaxy(redshift=0.5).has_mass_profile is False
        assert (
            aast.Galaxy(
                redshift=0.5, light_profile=aast.lp.LightProfile()
            ).has_mass_profile
            is False
        )
        assert (
            aast.Galaxy(
                redshift=0.5, mass_profile=aast.mp.MassProfile()
            ).has_mass_profile
            is True
        )

    def test_has_redshift(self):
        assert aast.Galaxy(redshift=0.1).has_redshift is True

    def test_has_pixelization(self):
        assert aast.Galaxy(redshift=0.5).has_pixelization is False
        assert (
            aast.Galaxy(
                redshift=0.5, pixelization=object(), regularization=object()
            ).has_pixelization
            is True
        )

    def test_has_regularization(self):
        assert aast.Galaxy(redshift=0.5).has_regularization is False
        assert (
            aast.Galaxy(
                redshift=0.5, pixelization=object(), regularization=object()
            ).has_regularization
            is True
        )

    def test_has_hyper_galaxy(self):
        assert aast.Galaxy(redshift=0.5, hyper_galaxy=object()).has_hyper_galaxy is True

    def test__only_pixelization_raises_error(self):
        with pytest.raises(exc.GalaxyException):
            aast.Galaxy(redshift=0.5, pixelization=object())

    def test__only_regularization_raises_error(self):
        with pytest.raises(exc.GalaxyException):
            aast.Galaxy(redshift=0.5, regularization=object())
